"""Pipelines package.

See https://github.com/BLSQ/openhexa/wiki/User-manual#using-pipelines and
https://github.com/BLSQ/openhexa/wiki/Writing-OpenHEXA-pipelines for more information about OpenHEXA pipelines.
"""

from .parameter import parameter
from .pipeline import Pipeline, pipeline
from .run import current_run
from .runtime import download_pipeline, import_pipeline
from .utils import get_local_workspace_config

__all__ = [
    "pipeline",
    "Pipeline",
    "parameter",
    "current_run",
    "current_pipeline",
    "import_pipeline",
    "download_pipeline",
    "get_local_workspace_config",
]



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import RemoveFromFavoritesError


class RemoveWebappFromFavorites(BaseModel):
    remove_from_favorites: "RemoveWebappFromFavoritesRemoveFromFavorites" = Field(
        alias="removeFromFavorites"
    )


class RemoveWebappFromFavoritesRemoveFromFavorites(BaseModel):
    success: bool
    errors: List[RemoveFromFavoritesError]


RemoveWebappFromFavorites.model_rebuild()



"""Workspace test module."""

import os
import re
from dataclasses import make_dataclass
from tempfile import mkdtemp
from unittest import mock

import pytest

from openhexa.graphql import GetConnectionConnectionBySlug
from openhexa.sdk.workspaces.connection import (
    CustomConnection,
    DHIS2Connection,
    GCSConnection,
    IASOConnection,
    PostgreSQLConnection,
    S3Connection,
)
from openhexa.sdk.workspaces.current_workspace import ConnectionDoesNotExist
from openhexa.utils import stringcase


class TestWorkspace:
    """Test class for Workspace."""

    def test_workspace_files_path(self, monkeypatch, workspace):
        """Basic checks for the Workspace.files_path() method."""
        assert workspace.files_path == "/home/hexa/workspace"

        monkeypatch.setenv("WORKSPACE_FILES_PATH", "/Users/John/openhexa/project-1/workspace")
        assert workspace.files_path == "/Users/John/openhexa/project-1/workspace"

    def test_workspace_tmp_path(self, monkeypatch, workspace):
        """Basic checks for the Workspace.tmp_path() method."""
        assert workspace.tmp_path == "/home/hexa/tmp"

        mock_tmp_path = mkdtemp()
        monkeypatch.setenv("WORKSPACE_TMP_PATH", mock_tmp_path)
        assert workspace.tmp_path == mock_tmp_path

    def test_workspace_dhis2_connection_not_exist(self, workspace):
        """Does not exist test case for DHIS2 connections."""
        identifier = "polio-ff3a0d"
        with pytest.raises(ConnectionDoesNotExist):
            workspace.dhis2_connection(identifier=identifier)

    def test_workspace_dhis2_connection(self, workspace):
        """Base test case for DHIS2 connections."""
        identifier = "polio-ff3a0d"
        env_variable_prefix = stringcase.constcase(identifier)
        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"
        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "dhis2",
                f"{env_variable_prefix}_URL": url,
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
            },
        ):
            dhis2_connection = workspace.dhis2_connection(identifier=identifier)
            assert dhis2_connection.url == url
            assert dhis2_connection.username == username
            assert dhis2_connection.password == password
            assert re.search("password", repr(dhis2_connection), re.IGNORECASE) is None
            assert re.search("password", str(dhis2_connection), re.IGNORECASE) is None

    def test_workspace_dhis2_connection_similar_prefix(self, workspace):
        """Base test case for DHIS2 connections."""
        identifier = "polio"
        identifier_2 = "polio-test"

        env_variable_prefix = stringcase.constcase(identifier)
        env_variable_prefix_2 = stringcase.constcase(identifier_2)

        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"

        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "dhis2",
                f"{env_variable_prefix}_URL": url,
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
                f"{env_variable_prefix_2}": "dhis2",
                f"{env_variable_prefix_2}_URL": "url_2",
                f"{env_variable_prefix_2}_USERNAME": "username_2",
                f"{env_variable_prefix_2}_PASSWORD": "password_2",
            },
        ):
            dhis2_connection = workspace.dhis2_connection(identifier=identifier)
            assert dhis2_connection.url == url
            assert dhis2_connection.username == username
            assert dhis2_connection.password == password

    def test_workspace_postgresql_connection_not_exist(self, workspace):
        """Does not exist test case for PostgreSQL connections."""
        identifier = "polio-ff3a0d"
        with pytest.raises(ConnectionDoesNotExist):
            workspace.postgresql_connection(identifier=identifier)

    def test_workspace_postgresql_connection(self, workspace):
        """Base test case for PostgreSQL connections."""
        identifier = "polio-ff3a0d"
        env_variable_prefix = stringcase.constcase(identifier)
        host = "https://172.17.0.1"
        port = "5432"
        username = "dhis2"
        password = "dhis2_pwd"
        database_name = "polio"
        url = f"postgresql://{username}:{password}@{host}:{port}/{database_name}"
        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "postgresql",
                f"{env_variable_prefix}_HOST": host,
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
                f"{env_variable_prefix}_PORT": port,
                f"{env_variable_prefix}_DB_NAME": database_name,
            },
        ):
            postgres_connection = workspace.postgresql_connection(identifier=identifier)
            assert postgres_connection.username == username
            assert postgres_connection.password == password
            assert postgres_connection.host == host
            assert postgres_connection.port == int(port)
            assert postgres_connection.database_name == database_name
            assert postgres_connection.url == url
            assert re.search("password", repr(postgres_connection), re.IGNORECASE) is None
            assert re.search("password", str(postgres_connection), re.IGNORECASE) is None

    def test_workspace_S3_connection_not_exist(self, workspace):
        """Does not exist test case for S3 connections."""
        identifier = "polio-ff3a0d"
        with pytest.raises(ConnectionDoesNotExist):
            workspace.s3_connection(identifier=identifier)

    def test_workspace_s3_connection(self, workspace):
        """Base test case for S3 connections."""
        identifier = "polio-ff3a0d"
        env_variable_prefix = stringcase.constcase(identifier)
        secret_access_key = "HqQBxH0BAI3zF7kANUNlGg"
        access_key_id = "84hVntMaMSYP/RSW9ex04w"
        bucket_name = "test"

        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "s3",
                f"{env_variable_prefix}_ACCESS_KEY_SECRET": secret_access_key,
                f"{env_variable_prefix}_ACCESS_KEY_ID": access_key_id,
                f"{env_variable_prefix}_BUCKET_NAME": bucket_name,
            },
        ):
            s3_connection = workspace.s3_connection(identifier=identifier)
            assert s3_connection.secret_access_key == secret_access_key
            assert s3_connection.access_key_id == access_key_id
            assert s3_connection.bucket_name == bucket_name
            assert re.search("secret_access_key", repr(s3_connection), re.IGNORECASE) is None
            assert re.search("secret_access_key", str(s3_connection), re.IGNORECASE) is None
            assert re.search("access_key_id", repr(s3_connection), re.IGNORECASE) is None
            assert re.search("access_key_id", str(s3_connection), re.IGNORECASE) is None

    def test_workspace_gcs_connection_not_exist(self, workspace):
        """Does not exist test case for GCS connections."""
        identifier = "polio-ff3a0d"
        with pytest.raises(ConnectionDoesNotExist):
            workspace.gcs_connection(identifier=identifier)

    def test_workspace_gcs_connection(self, workspace):
        """Base test case for GCS connections."""
        identifier = "polio-ff3a0d"
        env_variable_prefix = stringcase.constcase(identifier)
        service_account_key = "HqQBxH0BAI3zF7kANUNlGg"
        bucket_name = "test"

        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "gcs",
                f"{env_variable_prefix}_SERVICE_ACCOUNT_KEY": service_account_key,
                f"{env_variable_prefix}_BUCKET_NAME": bucket_name,
            },
        ):
            s3_connection = workspace.gcs_connection(identifier=identifier)
            assert s3_connection.service_account_key == service_account_key
            assert s3_connection.bucket_name == bucket_name
            assert re.search("service_account_key", repr(s3_connection), re.IGNORECASE) is None
            assert re.search("service_account_key", str(s3_connection), re.IGNORECASE) is None

    def test_workspace_iaso_connection_not_exist(self, workspace):
        """Does not exist test case for IASO connections."""
        identifier = "iaso-account"
        with pytest.raises(ConnectionDoesNotExist):
            workspace.iaso_connection(identifier=identifier)

    def test_workspace_iaso_connection(self, workspace):
        """Base test case for IASO connections."""
        identifier = "iaso-account"
        env_variable_prefix = stringcase.constcase(identifier)
        url = "https://test.iaso.org/"
        username = "iaso"
        password = "iaso_pwd"

        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "iaso",
                f"{env_variable_prefix}_URL": url,
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
            },
        ):
            iaso_connection = workspace.iaso_connection(identifier=identifier)
            assert iaso_connection.url == url
            assert iaso_connection.username == username
            assert iaso_connection.password == password
            assert re.search("password", repr(iaso_connection), re.IGNORECASE) is None
            assert re.search("password", str(iaso_connection), re.IGNORECASE) is None

    def test_workspace_custom_connection_not_exist(self, workspace):
        """Does not exist test case for custom connections."""
        identifier = "custom-stuff"
        with pytest.raises(ConnectionDoesNotExist):
            workspace.custom_connection(identifier=identifier)

    def test_workspace_custom_connection(self, workspace):
        """Base test case for custom connections."""
        identifier = "my_connection"
        env_variable_prefix = stringcase.constcase(identifier)
        username = "kaggle_username"
        password = "root"

        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "custom",
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
            },
        ):
            custom_connection = workspace.custom_connection(identifier=identifier)
            assert custom_connection.username == username
            assert custom_connection.password == password
            assert re.search("username", repr(custom_connection), re.IGNORECASE) is None
            assert re.search("password", str(custom_connection), re.IGNORECASE) is None

    def test_connection_by_slug_warning(self, workspace):
        """Ensure that using the slug keyword argument when getting a connection generates a deprecation warning."""
        identifier = "polio-ff3a0d"
        env_variable_prefix = stringcase.constcase(identifier)
        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"
        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "dhis2",
                f"{env_variable_prefix}_URL": url,
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
            },
        ):
            assert workspace.dhis2_connection(identifier).url == url
            with pytest.warns(DeprecationWarning):
                assert workspace.dhis2_connection(slug=identifier).url == url
            assert workspace.dhis2_connection(identifier=identifier).url == url

    def test_connection_various_case(self, workspace):
        """Ensure that identifiers used when getting connections are case-insensitive."""
        env_variable_prefix = stringcase.constcase("polio-123")
        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"
        with mock.patch.dict(
            os.environ,
            {
                f"{env_variable_prefix}": "dhis2",
                f"{env_variable_prefix}_URL": url,
                f"{env_variable_prefix}_USERNAME": username,
                f"{env_variable_prefix}_PASSWORD": password,
            },
        ):
            assert workspace.dhis2_connection("polio-123").url == url
            assert workspace.dhis2_connection("Polio-123").url == url
            assert workspace.dhis2_connection("POLIO-123").url == url


class TestConnectedWorkspace:
    """Test class for workspace that rely the graphl API."""

    @pytest.fixture(autouse=True)
    def setup_method(self, monkeypatch):
        """Basics setup for the TestConnectedWorkspace."""
        # Mock the class variable
        monkeypatch.setenv("HEXA_SERVER_URL", "http://app.openhexa.test")
        monkeypatch.setenv("HEXA_WORKSPACE", "workspace")

    def test_workspace_files_path(self, monkeypatch, workspace):
        """Basic checks for the Workspace.files_path() method."""
        assert workspace.files_path == "/home/hexa/workspace"

        monkeypatch.setenv("WORKSPACE_FILES_PATH", "/Users/John/openhexa/project-1/workspace")
        assert workspace.files_path == "/Users/John/openhexa/project-1/workspace"

    def test_workspace_tmp_path(self, monkeypatch, workspace):
        """Basic checks for the Workspace.tmp_path() method."""
        assert workspace.tmp_path == "/home/hexa/tmp"

        mock_tmp_path = mkdtemp()
        monkeypatch.setenv("WORKSPACE_TMP_PATH", mock_tmp_path)
        assert workspace.tmp_path == mock_tmp_path

    def test_workspace_get_connection_not_exist(self, workspace):
        """Test get connection not found."""
        data = None

        with mock.patch(
            "openhexa.sdk.workspaces.current_workspace.OpenHexaClient.get_connection",
            return_value=data,
        ):
            with pytest.raises(ValueError):
                workspace.get_connection("random")

    def test_workspace_get_connection_case_insensitive(self, workspace):
        """Test get connection."""
        data = {
            "__typename": "CustomConnection",
            "type": "CUSTOM",
            "fields": [{"code": "field_1", "value": "field_1_value"}],
        }
        mocked_data = GetConnectionConnectionBySlug(**data)
        with mock.patch(
            "openhexa.sdk.workspaces.current_workspace.OpenHexaClient.get_connection",
            return_value=mocked_data,
        ):
            connection = workspace.get_connection("RaNDom")
            assert isinstance(connection, CustomConnection)

    def test_workspace_get_connection(self, workspace):
        """Test get connection."""
        data = {
            "__typename": "S3Connection",
            "type": "S3",
            "fields": [
                {"code": "bucket_name", "value": "bucket_name"},
                {"code": "access_key_id", "value": "access_key_id"},
                {"code": "access_key_secret", "value": "secret_access_key"},
            ],
        }
        mocked_data = GetConnectionConnectionBySlug(**data)
        with mock.patch(
            "openhexa.sdk.workspaces.current_workspace.OpenHexaClient.get_connection",
            return_value=mocked_data,
        ):
            connection = workspace.get_connection("s3-connection")
            assert isinstance(connection, S3Connection)

    def test_workspace_dhis2_connection_not_exist(self, workspace):
        """Does not exist test case for DHIS2 connections."""
        identifier = "dhis2-play"
        with mock.patch.object(workspace, "get_connection", side_effect=ValueError()):
            with pytest.raises(ConnectionDoesNotExist):
                workspace.dhis2_connection(identifier=identifier)

    def test_workspace_dhis2_connection(self, workspace):
        """Base test case for DHIS2 connections."""
        identifier = "dhis2-play"
        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"

        data = DHIS2Connection(url, username, password)

        with mock.patch.object(workspace, "get_connection", return_value=data):
            dhis2_connection = workspace.dhis2_connection(identifier=identifier)

            assert dhis2_connection.url == url
            assert dhis2_connection.username == username
            assert dhis2_connection.password == password
            assert re.search("password", repr(dhis2_connection), re.IGNORECASE) is None
            assert re.search("password", str(dhis2_connection), re.IGNORECASE) is None

    def test_workspace_postgres_connection_not_exist(self, workspace):
        """Does not exist test case for Postgres connections."""
        identifier = "pg_data_play"

        with mock.patch.object(workspace, "get_connection", side_effect=ValueError()):
            with pytest.raises(ConnectionDoesNotExist):
                workspace.postgresql_connection(identifier=identifier)

    def test_workspace_postgresql_connection(self, workspace):
        """Base test case for PostgreSQL connections."""
        identifier = "polio-ff3a0d"

        host = "https://172.17.0.1"
        port = 5432
        username = "dhis2"
        password = "dhis2_pwd"
        database_name = "polio"
        url = f"postgresql://{username}:{password}@{host}:{port}/{database_name}"

        data = PostgreSQLConnection(host, port, username, password, database_name)
        with mock.patch.object(workspace, "get_connection", return_value=data):
            postgres_connection = workspace.postgresql_connection(identifier=identifier)
            assert postgres_connection.username == username
            assert postgres_connection.password == password
            assert postgres_connection.host == host
            assert postgres_connection.port == port
            assert postgres_connection.database_name == database_name
            assert postgres_connection.url == url
            assert re.search("password", repr(postgres_connection), re.IGNORECASE) is None
            assert re.search("password", str(postgres_connection), re.IGNORECASE) is None

    def test_workspace_S3_connection_not_exist(self, workspace):
        """Does not exist test case for S3 connections."""
        identifier = "polio-ff3a0d"
        with mock.patch.object(workspace, "get_connection", side_effect=ValueError()):
            with pytest.raises(ConnectionDoesNotExist):
                workspace.s3_connection(identifier=identifier)

    def test_workspace_s3_connection(monkeypatch, workspace):
        """Base test case for S3 connections."""
        identifier = "polio-ff3a0d"

        secret_access_key = "HqQBxH0BAI3zF7kANUNlGg"
        access_key_id = "84hVntMaMSYP/RSW9ex04w"
        bucket_name = "test"

        data = S3Connection(access_key_id, secret_access_key, bucket_name)
        with mock.patch.object(workspace, "get_connection", return_value=data):
            s3_connection = workspace.s3_connection(identifier=identifier)

            assert s3_connection.secret_access_key == secret_access_key
            assert s3_connection.access_key_id == access_key_id
            assert s3_connection.bucket_name == bucket_name
            assert re.search("secret_access_key", repr(s3_connection), re.IGNORECASE) is None
            assert re.search("secret_access_key", str(s3_connection), re.IGNORECASE) is None
            assert re.search("access_key_id", repr(s3_connection), re.IGNORECASE) is None
            assert re.search("access_key_id", str(s3_connection), re.IGNORECASE) is None

    def test_workspace_gcs_connection_not_exist(self, workspace):
        """Does not exist test case for GCS connections."""
        identifier = "polio-ff3a0d"
        with mock.patch.object(workspace, "get_connection", side_effect=ValueError()):
            with pytest.raises(ConnectionDoesNotExist):
                workspace.gcs_connection(identifier=identifier)

    def test_workspace_gcs_connection(self, workspace):
        """Base test case for GCS connections."""
        identifier = "polio-ff3a0d"

        service_account_key = "HqQBxH0BAI3zF7kANUNlGg"
        bucket_name = "test"

        data = GCSConnection(service_account_key, bucket_name)
        with mock.patch.object(workspace, "get_connection", return_value=data):
            s3_connection = workspace.gcs_connection(identifier=identifier)
            assert s3_connection.service_account_key == service_account_key
            assert s3_connection.bucket_name == bucket_name
            assert re.search("service_account_key", repr(s3_connection), re.IGNORECASE) is None
            assert re.search("service_account_key", str(s3_connection), re.IGNORECASE) is None

    def test_workspace_iaso_connection_not_exist(self, workspace):
        """Does not exist test case for IASO connections."""
        identifier = "iaso-account"
        with mock.patch.object(workspace, "get_connection", side_effect=ValueError()):
            with pytest.raises(ConnectionDoesNotExist):
                workspace.iaso_connection(identifier=identifier)

    def test_workspace_iaso_connection(self, workspace):
        """Base test case for IASO connections."""
        identifier = "iaso-account"

        url = "https://test.iaso.org/"
        username = "iaso"
        password = "iaso_pwd"

        data = IASOConnection(url, username, password)
        with mock.patch.object(workspace, "get_connection", return_value=data):
            iaso_connection = workspace.iaso_connection(identifier=identifier)
            assert iaso_connection.url == url
            assert iaso_connection.username == username
            assert iaso_connection.password == password
            assert re.search("password", repr(iaso_connection), re.IGNORECASE) is None
            assert re.search("password", str(iaso_connection), re.IGNORECASE) is None

    def test_workspace_custom_connection_not_exist(self, workspace):
        """Does not exist test case for custom connections."""
        identifier = "custom-stuff"
        with mock.patch.object(workspace, "get_connection", side_effect=ValueError()):
            with pytest.raises(ConnectionDoesNotExist):
                workspace.custom_connection(identifier=identifier)

    def test_workspace_custom_connection(self, workspace):
        """Base test case for custom connections."""
        identifier = "my_connection"

        username = "kaggle_username"
        password = "root"

        dataclass = make_dataclass(
            stringcase.pascalcase(identifier),
            ["username", "password"],
            bases=(CustomConnection,),
            repr=False,
        )
        with mock.patch.object(workspace, "get_connection", return_value=dataclass(username, password)):
            custom_connection = workspace.custom_connection(identifier=identifier)
            assert custom_connection.username == username
            assert custom_connection.password == password
            assert re.search("username", repr(custom_connection), re.IGNORECASE) is None
            assert re.search("password", str(custom_connection), re.IGNORECASE) is None

    def test_connection_by_slug_warning(self, workspace):
        """Ensure that using the slug keyword argument when getting a connection generates a deprecation warning."""
        identifier = "polio-ff3a0d"

        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"

        data = DHIS2Connection(url, username, password)
        with mock.patch.object(workspace, "get_connection", return_value=data):
            assert workspace.dhis2_connection(identifier).url == url
            with pytest.warns(DeprecationWarning):
                assert workspace.dhis2_connection(slug=identifier).url == url
            assert workspace.dhis2_connection(identifier=identifier).url == url

    def test_connection_various_case(self, workspace):
        """Ensure that identifiers used when getting connections are case-insensitive."""
        url = "https://test.dhis2.org/"
        username = "dhis2"
        password = "dhis2_pwd"
        data = DHIS2Connection(url, username, password)
        with mock.patch.object(workspace, "get_connection", return_value=data):
            assert workspace.dhis2_connection("polio-123").url == url
            assert workspace.dhis2_connection("Polio-123").url == url
            assert workspace.dhis2_connection("POLIO-123").url == url

    def test_workspace_configuration(self, workspace):
        mock_config = {
            "api_url": "https://api.example.com",
            "debug_mode": "true",
            "database_config": {"host": "localhost", "port": 5432},
        }

        mock_workspace_data = mock.Mock()
        mock_workspace_data.configuration = mock_config

        with mock.patch("openhexa.sdk.workspaces.current_workspace.OpenHexaClient") as mock_client:
            mock_client.return_value.workspace.return_value = mock_workspace_data
            assert workspace.configuration == mock_config

    def test_workspace_configuration_setter(self, workspace):
        """Test setting workspace configuration through property setter."""
        new_config = {
            "api_url": "https://new-api.example.com",
            "environment": "development",
            "features": {"feature1": True, "feature2": False},
        }

        mock_result = mock.Mock()
        mock_result.success = True

        with mock.patch("openhexa.sdk.workspaces.current_workspace.OpenHexaClient") as mock_client:
            mock_client.return_value.update_workspace.return_value = mock_result

            workspace.configuration = new_config

            mock_client.return_value.update_workspace.assert_called_once()
            call_args = mock_client.return_value.update_workspace.call_args[1]["input"]
            assert call_args.slug == workspace.slug
            assert call_args.configuration == new_config



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import InviteWorkspaceMembershipError


class InviteWorkspaceMember(BaseModel):
    invite_workspace_member: "InviteWorkspaceMemberInviteWorkspaceMember" = Field(
        alias="inviteWorkspaceMember"
    )


class InviteWorkspaceMemberInviteWorkspaceMember(BaseModel):
    success: bool
    errors: List[InviteWorkspaceMembershipError]
    workspace_membership: Optional[
        "InviteWorkspaceMemberInviteWorkspaceMemberWorkspaceMembership"
    ] = Field(alias="workspaceMembership")


class InviteWorkspaceMemberInviteWorkspaceMemberWorkspaceMembership(BaseModel):
    id: Any


InviteWorkspaceMember.model_rebuild()
InviteWorkspaceMemberInviteWorkspaceMember.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import CreatePipelineTemplateVersionError


class CreatePipelineTemplateVersion(BaseModel):
    create_pipeline_template_version: (
        "CreatePipelineTemplateVersionCreatePipelineTemplateVersion"
    ) = Field(alias="createPipelineTemplateVersion")


class CreatePipelineTemplateVersionCreatePipelineTemplateVersion(BaseModel):
    success: bool
    errors: Optional[List[CreatePipelineTemplateVersionError]]
    pipeline_template: Optional[
        "CreatePipelineTemplateVersionCreatePipelineTemplateVersionPipelineTemplate"
    ] = Field(alias="pipelineTemplate")


class CreatePipelineTemplateVersionCreatePipelineTemplateVersionPipelineTemplate(
    BaseModel
):
    name: str
    code: str


CreatePipelineTemplateVersion.model_rebuild()
CreatePipelineTemplateVersionCreatePipelineTemplateVersion.model_rebuild()



"""Miscellaneous utility functions."""

import abc
import contextlib
import enum
import os
import typing

from openhexa.graphql import BaseOpenHexaClient
from openhexa.utils import create_requests_session


class Environment(enum.Enum):
    """Enumeration of supported runtime environments."""

    LOCAL_PIPELINE = "LOCAL_PIPELINE"
    CLOUD_PIPELINE = "CLOUD_PIPELINE"
    CLOUD_JUPYTER = "CLOUD_JUPYTER"
    STANDALONE = "STANDALONE"


def get_environment():
    """Get the environment from the HEXA_ENVIRONMENT (see Environment enum)."""
    env = os.environ.get("HEXA_ENVIRONMENT", "STANDALONE").upper()

    try:
        return Environment[env]
    except KeyError:
        raise ValueError(f"Invalid environment: {env}")


def graphql(operation: str, variables: dict[str | typing.Any] | None = None) -> dict[str | typing.Any]:
    """Performa GraphQL query."""
    auth_token = os.environ[
        "HEXA_TOKEN"
    ]  # Works for notebooks with the membership token and pipelines with the run token
    headers = {"Authorization": f"Bearer {auth_token}"}
    session = create_requests_session()

    req = session.post(
        f"{os.environ['HEXA_SERVER_URL'].rstrip('/')}/graphql/",
        headers=headers,
        json={
            "query": operation,
            "variables": variables if variables is not None else {},
        },
    )
    req.raise_for_status()
    body = req.json()
    if "errors" in body:
        raise Exception(body["errors"])

    return body["data"]


class OpenHexaClient(BaseOpenHexaClient):
    """OpenHexaClient is a class that provides methods to interact with the OpenHexa GraphQL API."""

    def __init__(self, token: str | None = None, server_url: str | None = None):
        """Initialize the OpenHexaClient with the OpenHexa API URL and headers.

        Args:
            token: Authentication token. If not provided, will use HEXA_TOKEN environment variable.
            server_url: Server URL. If not provided, will use HEXA_SERVER_URL environment variable.
        """
        url = server_url or f"{os.environ['HEXA_SERVER_URL'].rstrip('/')}/graphql/"
        token = token or os.environ.get("HEXA_TOKEN")

        super().__init__(url=url, token=token)


class Iterator(metaclass=abc.ABCMeta):
    """A generic class for iterating through API list responses."""

    def __init__(
        self,
        item_to_value=lambda x: x,
        per_page=None,
    ):
        self._started = False
        self.__active_iterator = None

        self.item_to_value = item_to_value
        """Callable[Iterator, Any]: Callable to convert an item from the type
            in the raw API response into the native object. Will be called with
            the iterator and a
            single item.
        """
        self.per_page = per_page

        # The attributes below will change over the life of the iterator.
        self.page_number = 0
        """int: The current page of results."""
        self.num_results = 0
        """int: The total number of results fetched so far."""

    def _items_iter(self):
        for page in self._page_iter(increment=False):
            for item in page:
                self.num_results += 1
                yield item

    def __iter__(self) -> typing.Generator[typing.Any, None, None]:
        """Implement __iter()."""
        if self._started:
            raise ValueError("Iterator has already started", self)
        self._started = True

        return self._items_iter()

    def __next__(self):
        """Implement next()."""
        if self.__active_iterator is None:
            self.__active_iterator = iter(self)

        return next(self.__active_iterator)

    def _page_iter(self, increment: bool):
        """Generate pages of API responses.

        Parameters
        ----------
        increment : bool
            Flag indicating if the total number of results should be incremented on each page.
            This is useful since a page iterator will want to increment by results per page while an
            items iterator will want to increment per item.
        """
        page = self._next_page()
        while page is not None:
            self.page_number += 1
            if increment:
                self.num_results += page.num_items
            yield page
            page = self._next_page()

    @abc.abstractmethod
    def _next_page(self):
        """Get the next page in the iterator.

        This does nothing and is intended to be over-ridden by subclasses
        to return the next :class:`Page`.

        Raises
        ------
            NotImplementedError: Always, this method is abstract.
        """
        raise NotImplementedError


class Page:
    """Single page of results in an iterator.

    Parameters
    ----------
    parent : Iterator
        The iterator that owns the current page.
    items: Sequence[Any]
        An iterable (that also defines __len__) of items from a raw API response.
    item_to_value: Callable[dict[str, Any], Any]:
        Callable to convert an item from the type in the raw API response into the native object.
        Will be called with the iterator and a single item.
    """

    def __init__(self, parent, items, item_to_value):
        self._parent = parent
        self._num_items = len(items)
        self._remaining = self._num_items
        self._item_iter = iter(items)
        self._item_to_value = item_to_value

    @property
    def num_items(self):
        """int: Total items in the page."""
        return self._num_items

    @property
    def remaining(self):
        """int: Remaining items in the page."""
        return self._remaining

    def __iter__(self):
        """Implement __iter__()."""
        return self

    def __next__(self):
        """Get the next value in the page."""
        item = next(self._item_iter)
        result = self._item_to_value(item)
        # Since we've successfully got the next value from the
        # iterator, we update the number of remaining.
        self._remaining -= 1
        return result


@contextlib.contextmanager
def read_content(source: str | os.PathLike[str] | typing.IO | bytes):
    """Read file content and return it as bytes."""
    try:
        if isinstance(source, bytes):
            yield source
        elif hasattr(source, "read"):
            yield source
        # If source is a string or PathLike object
        elif isinstance(source, (str | os.PathLike)):
            with open(os.fspath(source), "rb") as f:
                yield f

        else:
            raise ValueError("Unsupported type for source")
    finally:
        if hasattr(source, "close"):
            source.close()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import CreateWebappError


class CreateWebapp(BaseModel):
    create_webapp: "CreateWebappCreateWebapp" = Field(alias="createWebapp")


class CreateWebappCreateWebapp(BaseModel):
    success: bool
    errors: List[CreateWebappError]
    webapp: Optional["CreateWebappCreateWebappWebapp"]


class CreateWebappCreateWebappWebapp(BaseModel):
    id: Any


CreateWebapp.model_rebuild()
CreateWebappCreateWebapp.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List, Literal, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import UpdateConnectionError


class UpdateConnection(BaseModel):
    update_connection: "UpdateConnectionUpdateConnection" = Field(
        alias="updateConnection"
    )


class UpdateConnectionUpdateConnection(BaseModel):
    success: bool
    errors: List[UpdateConnectionError]
    connection: Optional["UpdateConnectionUpdateConnectionConnection"]


class UpdateConnectionUpdateConnectionConnection(BaseModel):
    typename__: Literal[
        "Connection",
        "CustomConnection",
        "DHIS2Connection",
        "GCSConnection",
        "IASOConnection",
        "PostgreSQLConnection",
        "S3Connection",
    ] = Field(alias="__typename")
    id: str
    name: str
    slug: str
    description: Optional[str]
    fields: List["UpdateConnectionUpdateConnectionConnectionFields"]


class UpdateConnectionUpdateConnectionConnectionFields(BaseModel):
    code: str
    value: Optional[str]
    secret: bool


UpdateConnection.model_rebuild()
UpdateConnectionUpdateConnection.model_rebuild()
UpdateConnectionUpdateConnectionConnection.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import PipelineError


class UploadPipeline(BaseModel):
    upload_pipeline: "UploadPipelineUploadPipeline" = Field(alias="uploadPipeline")


class UploadPipelineUploadPipeline(BaseModel):
    success: bool
    errors: List[PipelineError]


UploadPipeline.model_rebuild()



"""Module-level test fixtures."""

from importlib import reload
from unittest import mock

import pytest

import openhexa.cli
import openhexa.sdk
from openhexa.sdk.pipelines.log_level import LogLevel


@pytest.fixture(scope="function")
def workspace():
    """Build workspace fixture."""
    from openhexa.sdk import workspace as global_workspace

    reload(openhexa.sdk)

    return global_workspace


@pytest.fixture(scope="function")
def current_run():
    """Build current run fixture."""
    from openhexa.sdk import current_run as global_current_run

    reload(openhexa.sdk)

    return global_current_run


@pytest.fixture(scope="function", autouse=True)
def settings(monkeypatch):
    """Build settings fixture."""
    settings_mock = mock.MagicMock()
    monkeypatch.setattr("openhexa.cli.settings.settings", settings_mock)
    reload(openhexa.cli.api)
    settings_mock.current_workspace = "workspace-slug"
    settings_mock.api_url = "http://localhost:8000/graphql"
    settings_mock.workspaces = {"workspace-slug": "token", "another-workspace-slug": "token"}
    settings_mock.debug = False
    settings_mock.access_token = "token"
    settings_mock.log_level = LogLevel.INFO
    return settings_mock



# Generated by ariadne-codegen
# Source: openhexa/graphql/schema.generated.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import (
    AccessmodAccessibilityAnalysisAlgorithm,
    ConnectionType,
    MembershipRole,
    MessagePriority,
    ParameterWidget,
    PermissionMode,
    PipelineNotificationLevel,
    WorkspaceMembershipRole,
)


class AddPipelineOutputInput(BaseModel):
    name: Optional[str] = None
    type: str
    uri: str


class AddToFavoritesInput(BaseModel):
    webapp_id: str = Field(alias="webappId")


class ApproveAccessmodAccessRequestInput(BaseModel):
    id: str


class ArchiveWorkspaceInput(BaseModel):
    slug: str


class ConnectionFieldInput(BaseModel):
    code: str
    secret: bool
    value: Optional[str] = None


class CountryInput(BaseModel):
    alpha_3: Optional[str] = Field(alias="alpha3", default=None)
    code: str
    flag: Optional[str] = None
    name: Optional[str] = None


class CreateAccessmodAccessibilityAnalysisInput(BaseModel):
    name: str
    project_id: str = Field(alias="projectId")


class CreateAccessmodFileInput(BaseModel):
    fileset_id: str = Field(alias="filesetId")
    mime_type: str = Field(alias="mimeType")
    uri: str


class CreateAccessmodFilesetInput(BaseModel):
    automatic: Optional[bool] = None
    metadata: Optional[Any] = None
    name: str
    project_id: str = Field(alias="projectId")
    role_id: str = Field(alias="roleId")


class CreateAccessmodProjectInput(BaseModel):
    country: "CountryInput"
    crs: int
    description: Optional[str] = None
    extent: Optional[List[List[float]]] = None
    name: str
    spatial_resolution: int = Field(alias="spatialResolution")


class CreateAccessmodProjectMemberInput(BaseModel):
    mode: PermissionMode
    project_id: str = Field(alias="projectId")
    team_id: Optional[str] = Field(alias="teamId", default=None)
    user_id: Optional[str] = Field(alias="userId", default=None)


class CreateAccessmodZonalStatisticsInput(BaseModel):
    name: str
    project_id: str = Field(alias="projectId")


class CreateBucketFolderInput(BaseModel):
    folder_key: str = Field(alias="folderKey")
    workspace_slug: str = Field(alias="workspaceSlug")


class CreateConnectionInput(BaseModel):
    description: Optional[str] = None
    fields: Optional[List["ConnectionFieldInput"]] = None
    name: str
    slug: Optional[str] = None
    type: ConnectionType
    workspace_slug: str = Field(alias="workspaceSlug")


class CreateDatasetInput(BaseModel):
    description: Optional[str] = None
    name: str
    workspace_slug: str = Field(alias="workspaceSlug")


class CreateDatasetVersionFileInput(BaseModel):
    content_type: str = Field(alias="contentType")
    uri: str
    version_id: str = Field(alias="versionId")


class CreateDatasetVersionInput(BaseModel):
    changelog: Optional[str] = None
    dataset_id: str = Field(alias="datasetId")
    name: str


class CreateMembershipInput(BaseModel):
    role: MembershipRole
    team_id: Any = Field(alias="teamId")
    user_email: str = Field(alias="userEmail")


class CreatePipelineFromTemplateVersionInput(BaseModel):
    pipeline_template_version_id: Any = Field(alias="pipelineTemplateVersionId")
    workspace_slug: str = Field(alias="workspaceSlug")


class CreatePipelineInput(BaseModel):
    code: Optional[str] = None
    name: str
    notebook_path: Optional[str] = Field(alias="notebookPath", default=None)
    workspace_slug: str = Field(alias="workspaceSlug")


class CreatePipelineRecipientInput(BaseModel):
    notification_level: PipelineNotificationLevel = Field(alias="notificationLevel")
    pipeline_id: Any = Field(alias="pipelineId")
    user_id: Any = Field(alias="userId")


class CreatePipelineTemplateVersionInput(BaseModel):
    changelog: Optional[str] = None
    code: Optional[str] = None
    config: Optional[str] = None
    description: Optional[str] = None
    name: Optional[str] = None
    pipeline_id: Any = Field(alias="pipelineId")
    pipeline_version_id: Any = Field(alias="pipelineVersionId")
    workspace_slug: str = Field(alias="workspaceSlug")


class CreateTeamInput(BaseModel):
    name: str


class CreateWebappInput(BaseModel):
    description: Optional[str] = None
    icon: Optional[str] = None
    name: str
    url: str
    workspace_slug: str = Field(alias="workspaceSlug")


class CreateWorkspaceInput(BaseModel):
    countries: Optional[List["CountryInput"]] = None
    description: Optional[str] = None
    load_sample_data: Optional[bool] = Field(alias="loadSampleData", default=None)
    name: str
    slug: Optional[str] = None


class DeclineWorkspaceInvitationInput(BaseModel):
    invitation_id: Any = Field(alias="invitationId")


class DeleteAccessmodAnalysisInput(BaseModel):
    id: str


class DeleteAccessmodFilesetInput(BaseModel):
    id: str


class DeleteAccessmodProjectInput(BaseModel):
    id: str


class DeleteAccessmodProjectMemberInput(BaseModel):
    id: str


class DeleteBucketObjectInput(BaseModel):
    object_key: str = Field(alias="objectKey")
    workspace_slug: str = Field(alias="workspaceSlug")


class DeleteConnectionInput(BaseModel):
    id: str


class DeleteDatasetInput(BaseModel):
    id: str


class DeleteDatasetLinkInput(BaseModel):
    id: str


class DeleteDatasetVersionInput(BaseModel):
    version_id: str = Field(alias="versionId")


class DeleteMembershipInput(BaseModel):
    id: Any


class DeleteMetadataAttributeInput(BaseModel):
    key: str
    target_id: Any = Field(alias="targetId")


class DeletePipelineInput(BaseModel):
    id: Any


class DeletePipelineRecipientInput(BaseModel):
    recipient_id: Any = Field(alias="recipientId")


class DeletePipelineTemplateInput(BaseModel):
    id: Any


class DeletePipelineVersionInput(BaseModel):
    id: Any


class DeleteTeamInput(BaseModel):
    id: Any


class DeleteTemplateVersionInput(BaseModel):
    id: Any


class DeleteWebappInput(BaseModel):
    id: Any


class DeleteWorkspaceDatabaseTableInput(BaseModel):
    table: str
    workspace_slug: str = Field(alias="workspaceSlug")


class DeleteWorkspaceInput(BaseModel):
    slug: str


class DeleteWorkspaceInvitationInput(BaseModel):
    invitation_id: Any = Field(alias="invitationId")


class DeleteWorkspaceMemberInput(BaseModel):
    membership_id: Any = Field(alias="membershipId")


class DenyAccessmodAccessRequestInput(BaseModel):
    id: str


class DisableTwoFactorInput(BaseModel):
    token: str


class EnableTwoFactorInput(BaseModel):
    email: Optional[str] = None


class GenerateDatasetUploadUrlInput(BaseModel):
    content_type: str = Field(alias="contentType")
    uri: str
    version_id: str = Field(alias="versionId")


class GenerateNewDatabasePasswordInput(BaseModel):
    workspace_slug: str = Field(alias="workspaceSlug")


class GeneratePipelineWebhookUrlInput(BaseModel):
    id: Any


class GenerateWorkspaceTokenInput(BaseModel):
    slug: str


class IASOQueryFilterInput(BaseModel):
    type: str
    value: List[int]


class InviteWorkspaceMemberInput(BaseModel):
    role: WorkspaceMembershipRole
    user_email: str = Field(alias="userEmail")
    workspace_slug: str = Field(alias="workspaceSlug")


class JoinWorkspaceInput(BaseModel):
    invitation_id: Any = Field(alias="invitationId")


class LaunchAccessmodAnalysisInput(BaseModel):
    id: str


class LaunchNotebookServerInput(BaseModel):
    workspace_slug: str = Field(alias="workspaceSlug")


class LinkDatasetInput(BaseModel):
    dataset_id: str = Field(alias="datasetId")
    workspace_slug: str = Field(alias="workspaceSlug")


class LogPipelineMessageInput(BaseModel):
    message: str
    priority: MessagePriority


class LoginInput(BaseModel):
    email: str
    password: str
    token: Optional[str] = None


class OrganizationInput(BaseModel):
    contact_info: Optional[str] = Field(alias="contactInfo", default=None)
    id: Any
    name: Optional[str] = None
    type: Optional[str] = None
    url: Optional[str] = None


class ParameterInput(BaseModel):
    choices: Optional[List[Any]] = None
    code: str
    connection: Optional[str] = None
    default: Optional[Any] = None
    help: Optional[str] = None
    multiple: Optional[bool] = None
    name: Optional[str] = None
    required: Optional[bool] = None
    type: str
    widget: Optional[ParameterWidget] = None


class PinDatasetInput(BaseModel):
    link_id: str = Field(alias="linkId")
    pinned: bool


class PipelineTokenInput(BaseModel):
    pipeline_code: str = Field(alias="pipelineCode")
    workspace_slug: str = Field(alias="workspaceSlug")


class PrepareAccessmodFileDownloadInput(BaseModel):
    file_id: str = Field(alias="fileId")


class PrepareAccessmodFileUploadInput(BaseModel):
    fileset_id: str = Field(alias="filesetId")
    mime_type: str = Field(alias="mimeType")


class PrepareAccessmodFilesetVisualizationDownloadInput(BaseModel):
    id: str


class PrepareDownloadURLInput(BaseModel):
    uri: Any


class PrepareObjectDownloadInput(BaseModel):
    object_key: str = Field(alias="objectKey")
    workspace_slug: str = Field(alias="workspaceSlug")


class PrepareObjectUploadInput(BaseModel):
    content_type: Optional[str] = Field(alias="contentType", default=None)
    object_key: str = Field(alias="objectKey")
    workspace_slug: str = Field(alias="workspaceSlug")


class PrepareVersionFileDownloadInput(BaseModel):
    file_id: str = Field(alias="fileId")


class RegisterInput(BaseModel):
    first_name: str = Field(alias="firstName")
    invitation_token: str = Field(alias="invitationToken")
    last_name: str = Field(alias="lastName")
    password_1: str = Field(alias="password1")
    password_2: str = Field(alias="password2")


class RemoveFromFavoritesInput(BaseModel):
    webapp_id: str = Field(alias="webappId")


class RequestAccessmodAccessInput(BaseModel):
    accept_tos: bool = Field(alias="acceptTos")
    email: str
    first_name: str = Field(alias="firstName")
    last_name: str = Field(alias="lastName")


class ResendWorkspaceInvitationInput(BaseModel):
    invitation_id: Any = Field(alias="invitationId")


class ResetPasswordInput(BaseModel):
    email: str


class RunDAGInput(BaseModel):
    config: Any
    dag_id: Any = Field(alias="dagId")


class RunPipelineInput(BaseModel):
    config: Any
    enable_debug_logs: Optional[bool] = Field(alias="enableDebugLogs", default=None)
    id: Any
    send_mail_notifications: Optional[bool] = Field(
        alias="sendMailNotifications", default=None
    )
    version_id: Optional[Any] = Field(alias="versionId", default=None)


class SetDAGRunFavoriteInput(BaseModel):
    id: Any
    is_favorite: bool = Field(alias="isFavorite")
    label: Optional[str] = None


class SetMetadataAttributeInput(BaseModel):
    key: str
    label: Optional[str] = None
    target_id: Any = Field(alias="targetId")
    value: Optional[Any] = None


class SetPasswordInput(BaseModel):
    password_1: str = Field(alias="password1")
    password_2: str = Field(alias="password2")
    token: str
    uidb_64: str = Field(alias="uidb64")


class StopPipelineInput(BaseModel):
    run_id: Any = Field(alias="runId")


class UpdateAccessmodAccessibilityAnalysisInput(BaseModel):
    algorithm: Optional[AccessmodAccessibilityAnalysisAlgorithm] = None
    barrier_id: Optional[str] = Field(alias="barrierId", default=None)
    dem_id: Optional[str] = Field(alias="demId", default=None)
    health_facilities_id: Optional[str] = Field(
        alias="healthFacilitiesId", default=None
    )
    id: str
    invert_direction: Optional[bool] = Field(alias="invertDirection", default=None)
    knight_move: Optional[bool] = Field(alias="knightMove", default=None)
    land_cover_id: Optional[str] = Field(alias="landCoverId", default=None)
    max_travel_time: Optional[int] = Field(alias="maxTravelTime", default=None)
    moving_speeds: Optional[Any] = Field(alias="movingSpeeds", default=None)
    name: Optional[str] = None
    stack_id: Optional[str] = Field(alias="stackId", default=None)
    stack_priorities: Optional[Any] = Field(alias="stackPriorities", default=None)
    transport_network_id: Optional[str] = Field(
        alias="transportNetworkId", default=None
    )
    water_all_touched: Optional[bool] = Field(alias="waterAllTouched", default=None)
    water_id: Optional[str] = Field(alias="waterId", default=None)


class UpdateAccessmodFilesetInput(BaseModel):
    id: str
    metadata: Optional[Any] = None
    name: Optional[str] = None


class UpdateAccessmodProjectInput(BaseModel):
    description: Optional[str] = None
    id: str
    name: Optional[str] = None


class UpdateAccessmodProjectMemberInput(BaseModel):
    id: str
    mode: PermissionMode


class UpdateAccessmodZonalStatisticsInput(BaseModel):
    boundaries_id: Optional[str] = Field(alias="boundariesId", default=None)
    id: str
    name: Optional[str] = None
    population_id: Optional[str] = Field(alias="populationId", default=None)
    time_thresholds: Optional[Any] = Field(alias="timeThresholds", default=None)
    travel_times_id: Optional[str] = Field(alias="travelTimesId", default=None)


class UpdateConnectionInput(BaseModel):
    description: Optional[str] = None
    fields: Optional[List["ConnectionFieldInput"]] = None
    id: str
    name: Optional[str] = None
    slug: Optional[str] = None


class UpdateDAGInput(BaseModel):
    countries: Optional[List["CountryInput"]] = None
    description: Optional[str] = None
    id: Any
    label: Optional[str] = None
    schedule: Optional[str] = None


class UpdateDatasetInput(BaseModel):
    dataset_id: str = Field(alias="datasetId")
    description: Optional[str] = None
    name: Optional[str] = None


class UpdateDatasetVersionInput(BaseModel):
    changelog: Optional[str] = None
    name: Optional[str] = None
    version_id: str = Field(alias="versionId")


class UpdateMembershipInput(BaseModel):
    id: Any
    role: MembershipRole


class UpdatePipelineInput(BaseModel):
    config: Optional[Any] = None
    description: Optional[str] = None
    id: Any
    name: Optional[str] = None
    schedule: Optional[str] = None
    webhook_enabled: Optional[bool] = Field(alias="webhookEnabled", default=None)


class UpdatePipelineProgressInput(BaseModel):
    percent: int


class UpdatePipelineRecipientInput(BaseModel):
    notification_level: PipelineNotificationLevel = Field(alias="notificationLevel")
    recipient_id: Any = Field(alias="recipientId")


class UpdatePipelineVersionInput(BaseModel):
    config: Optional[Any] = None
    description: Optional[str] = None
    external_link: Optional[Any] = Field(alias="externalLink", default=None)
    id: Any
    name: Optional[str] = None


class UpdateTeamInput(BaseModel):
    id: Any
    name: Optional[str] = None


class UpdateTemplateInput(BaseModel):
    config: Optional[Any] = None
    description: Optional[str] = None
    id: Any
    name: Optional[str] = None


class UpdateTemplateVersionInput(BaseModel):
    changelog: Optional[str] = None
    id: Any


class UpdateUserInput(BaseModel):
    first_name: Optional[str] = Field(alias="firstName", default=None)
    language: Optional[str] = None
    last_name: Optional[str] = Field(alias="lastName", default=None)


class UpdateWebappInput(BaseModel):
    description: Optional[str] = None
    icon: Optional[str] = None
    id: Any
    name: Optional[str] = None
    url: Optional[str] = None


class UpdateWorkspaceInput(BaseModel):
    configuration: Optional[Any] = None
    countries: Optional[List["CountryInput"]] = None
    description: Optional[str] = None
    docker_image: Optional[str] = Field(alias="dockerImage", default=None)
    name: Optional[str] = None
    slug: str


class UpdateWorkspaceMemberInput(BaseModel):
    membership_id: Any = Field(alias="membershipId")
    role: WorkspaceMembershipRole


class UpgradePipelineVersionFromTemplateInput(BaseModel):
    pipeline_id: Any = Field(alias="pipelineId")


class UploadPipelineInput(BaseModel):
    code: Optional[str] = None
    config: Optional[Any] = None
    description: Optional[str] = None
    external_link: Optional[Any] = Field(alias="externalLink", default=None)
    name: Optional[str] = None
    parameters: List["ParameterInput"]
    pipeline_code: Optional[str] = Field(alias="pipelineCode", default=None)
    timeout: Optional[int] = None
    workspace_slug: str = Field(alias="workspaceSlug")
    zipfile: str


class VerifyDeviceInput(BaseModel):
    token: Optional[str] = None


CreateAccessmodProjectInput.model_rebuild()
CreateConnectionInput.model_rebuild()
CreateWorkspaceInput.model_rebuild()
UpdateConnectionInput.model_rebuild()
UpdateDAGInput.model_rebuild()
UpdateWorkspaceInput.model_rebuild()
UploadPipelineInput.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import ArchiveWorkspaceError


class ArchiveWorkspace(BaseModel):
    archive_workspace: "ArchiveWorkspaceArchiveWorkspace" = Field(
        alias="archiveWorkspace"
    )


class ArchiveWorkspaceArchiveWorkspace(BaseModel):
    success: bool
    errors: List[ArchiveWorkspaceError]


ArchiveWorkspace.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List, Optional

from pydantic import Field

from .base_model import BaseModel


class Workspaces(BaseModel):
    workspaces: "WorkspacesWorkspaces"


class WorkspacesWorkspaces(BaseModel):
    total_pages: int = Field(alias="totalPages")
    items: List["WorkspacesWorkspacesItems"]


class WorkspacesWorkspacesItems(BaseModel):
    slug: str
    name: str
    description: Optional[str]
    countries: List["WorkspacesWorkspacesItemsCountries"]


class WorkspacesWorkspacesItemsCountries(BaseModel):
    code: str
    alpha_3: str = Field(alias="alpha3")
    name: str


Workspaces.model_rebuild()
WorkspacesWorkspaces.model_rebuild()
WorkspacesWorkspacesItems.model_rebuild()



"""Simple module for a sample logistic pipeline."""

import json
import typing
from io import BytesIO

import geopandas as gpd
import pandas as pd
import rasterio
import requests
from rasterstats import zonal_stats

from openhexa.sdk import parameter, pipeline, workspace


@pipeline("Logistic stats", "Test")
@parameter(
    "deg",
    name="Data element group",
    type=str,
    help="The ID of the data element group in DHIS2.",
)
@parameter(
    "periods",
    type=str,
    multiple=True,
    help="A list of DHIS2 periods, separated by a newline.",
)
@parameter("oul", name="Organisation unit level", type=int, default=2)
def logistic_stats(deg: str, periods: str, oul: int):
    """Run a basic logistic stats pipeline."""
    dhis2_data = dhis2_download(deg, periods, oul)
    gadm_data = gadm_download()
    worldpop_data = worldpop_download()

    model(dhis2_data, gadm_data, worldpop_data)


@logistic_stats.task
def dhis2_download(data_element_group: str, periods: str, org_unit_level: int) -> dict[str, typing.Any]:
    """Download DHIS2 data."""
    connection = workspace.dhis2_connection("dhis2-play")
    base_url = f"{connection.url}/api"
    session = requests.Session()
    session.auth = (connection.username, connection.password)

    # Store config in a static file
    dx = [f"DE_GROUP-{data_element_group}"]

    analytics_response = session.get(
        f"{base_url}/analytics",
        params={
            "aggregationType": "SUM",
            # "fields": "id",
            "dimension": [
                f"dx:{';'.join(dx)}",
                f"pe:{';'.join(periods)}",
                f"ou:LEVEL-{org_unit_level}",
            ],
        },
    )

    return json.loads(analytics_response.text)


@logistic_stats.task
def gadm_download():
    """Download administrative boundaries data from UCDavis."""
    url = "https://geodata.ucdavis.edu/gadm/gadm4.1/gpkg/gadm41_SLE.gpkg"
    r = requests.get(url, timeout=30)

    return r.content


@logistic_stats.task
def worldpop_download():
    """Download population data from worldpop.org."""
    base_url = "https://data.worldpop.org/"
    url = f"{base_url}GIS/Population/Global_2000_2020_Constrained/2020/maxar_v1/SLE/sle_ppp_2020_UNadj_constrained.tif"
    r = requests.get(url)

    return r.content


@logistic_stats.task
def model(dhis2_data: dict[str, typing.Any], gadm_data, worldpop_data):
    """Build a basic data model."""
    # Load DHIS2 data
    dhis2_df = pd.DataFrame(dhis2_data["rows"], columns=[h["column"] for h in dhis2_data["headers"]])
    dhis2_df = dhis2_df.rename(columns={"Data": "Data element id", "Organisation unit": "Organisation unit id"})
    dhis2_df["Value"] = pd.to_numeric(dhis2_df["Value"], downcast="integer")
    dhis2_df["Period"] = pd.PeriodIndex(dhis2_df["Period"], freq="Q")
    dhis2_df["Organisation unit"] = dhis2_df["Organisation unit id"].map(
        lambda ouid: dhis2_data["metaData"]["items"][ouid]["name"]
    )
    dhis2_df["Data element"] = dhis2_df["Data element id"].map(
        lambda deid: dhis2_data["metaData"]["items"][deid]["name"]
    )

    print(f"Load DHIS2 DF - {len(dhis2_df)} rows")

    # Load & compute population data
    administrative_areas = gpd.read_file(BytesIO(gadm_data), layer="ADM_ADM_2")
    with rasterio.open(BytesIO(worldpop_data)) as src:
        stats = zonal_stats(
            administrative_areas,  # the geometries
            src.read(1),  # the source raster data
            affine=src.meta["transform"],  # affine transform of the source raster
            stats=["sum"],  # the zonal stats we want to compute
            nodata=src.meta["nodata"],  # nodata value in the source raster
        )

    population_df = pd.Series(data=[stat["sum"] for stat in stats], index=administrative_areas.index)
    population_df = pd.DataFrame({"District": administrative_areas["NAME_2"], "Population": population_df})
    corrected_population_df = population_df.copy()
    corrected_population_df.loc[
        corrected_population_df["District"].str.startswith("Western"), "District"
    ] = "Western Area"

    corrected_population_df = corrected_population_df.groupby("District").sum()

    print(f"Built population DF - {len(population_df)} rows")

    # Compute stats
    pivot_df = dhis2_df.pivot(index=["Period", "Organisation unit"], columns="Data element", values="Value")
    combined_df = pivot_df.reset_index().merge(
        corrected_population_df,
        how="left",
        left_on="Organisation unit",
        right_index=True,
    )
    combined_df = combined_df.set_index(["Period", "Organisation unit"])
    # TODO: replace by something dynamic
    # combined_df["Oral Rehydration Salts by 100000 inh."] = combined_df[
    #     "Commodities - Oral Rehydration Salts"
    # ] / (combined_df["Population"] / 100000)
    # combined_df["Oxytocin by 100000 inh."] = combined_df["Commodities - Oxytocin"] / (
    #     combined_df["Population"] / 100000
    # )

    # Save to CSV

    # TODO: variable path / helper for workspace path
    combined_df.to_csv(f"{workspace.files_path}/stats.csv")


if __name__ == "__main__":
    logistic_stats()



"""Collection of functions that interacts with the OpenHEXA API."""

import base64
import enum
import io
import json
import logging
import os
import tempfile
import typing
from datetime import datetime
from importlib.metadata import version
from pathlib import Path
from typing import Any
from zipfile import ZipFile

import click
import docker
import httpx
import requests
from docker.models.containers import Container
from graphql import build_client_schema, build_schema, get_introspection_query
from graphql.utilities import find_breaking_changes
from jinja2 import Template

from openhexa.cli.settings import settings
from openhexa.graphql import BaseOpenHexaClient
from openhexa.sdk.pipelines import get_local_workspace_config
from openhexa.sdk.pipelines.runtime import get_pipeline
from openhexa.utils import create_requests_session, stringcase


class InvalidDefinitionError(Exception):
    """Raised whenever pipeline parameters and/or pipeline options are incompatible."""

    pass


class NoActiveWorkspaceError(Exception):
    """Raised when no workspace is currently active in the configuration."""

    pass


class OutputDirectoryError(Exception):
    """Raised when the output directory is not a directory or is not empty."""

    pass


class APIError(Exception):
    """Raised when an error occurs while interacting with the API."""

    pass


class InvalidTokenError(APIError):
    """Raised when the token is invalid."""

    pass


class GraphQLError(APIError):
    """Raised when a GraphQL request returns an error."""

    pass


class PipelineDirectoryError(Exception):
    """Raised when the pipeline directory is not a directory or does not exist."""

    pass


class DockerError(Exception):
    """Raised when Docker is not running or is not installed."""

    pass


class PipelineDefinitionErrorCode(enum.Enum):
    """Enumeration of possible pipeline definition error codes."""

    PIPELINE_DOES_NOT_SUPPORT_PARAMETERS = "PIPELINE_DOES_NOT_SUPPORT_PARAMETERS"
    INVALID_TIMEOUT_VALUE = "INVALID_TIMEOUT_VALUE"
    DUPLICATE_PIPELINE_VERSION_NAME = "DUPLICATE_PIPELINE_VERSION_NAME"


class PermissionDenied(Exception):
    """Raised whenever an operation on a pipeline is denied by the backend."""

    pass


def get_library_versions() -> tuple[str, str]:
    """Return the current version and the one on PyPi."""
    # Get the currently installed version
    installed_version = version("openhexa.sdk")

    # Get the latest version available on PyPI
    try:
        response = requests.get("https://pypi.org/pypi/openhexa.sdk/json")
        latest_version = response.json()["info"]["version"]
        return installed_version, latest_version
    except requests.RequestException:
        logging.error(
            "Could not check for the latest version of the openhexa.sdk package.",
            exc_info=True,
        )
        return installed_version, installed_version


def _detect_graphql_breaking_changes_if_needed(token):
    """Detect breaking changes if not done recently between the schema referenced in the SDK and the server using graphql-core."""
    ONE_HOUR = 60 * 60
    now_timestamp = int(datetime.now().timestamp())
    if not settings.last_breaking_change_check or now_timestamp - settings.last_breaking_change_check > ONE_HOUR:
        _detect_graphql_breaking_changes(token)
        settings.last_breaking_change_check = now_timestamp


def _detect_graphql_breaking_changes(token):
    """Detect breaking changes between the schema referenced in the SDK and the server using graphql-core."""
    stored_schema_obj = build_schema(
        (Path(__file__).parent.parent / "graphql" / "schema.generated.graphql").open().read()
    )
    server_schema_obj = build_client_schema(
        _query_graphql(get_introspection_query(input_value_deprecation=True), token=token)
    )

    breaking_changes = find_breaking_changes(stored_schema_obj, server_schema_obj)
    if breaking_changes:
        current_version, latest_version = get_library_versions()
        click.secho(
            f"⚠️ Breaking changes detected between the SDK (version {current_version}) and the server:",
            fg="red",
        )
        for change in breaking_changes:
            click.secho(f"- {change.description}", fg="yellow")
        click.secho(
            "This could lead to unexpected results.\n"
            f"Please update the SDK to the latest version {latest_version} "
            f"(using `pip install openhexa-sdk=={latest_version}`) or use a version of the SDK compatible with the server.",
            fg="red",
        )


def graphql(query: str, variables=None, token=None):
    """Check that there is no breaking change and perform a GraphQL request."""
    _detect_graphql_breaking_changes_if_needed(token)
    return _query_graphql(query, variables, token)


def _query_graphql(query: str, variables=None, token=None):
    """Perform a GraphQL request."""
    url = settings.api_url + "/graphql/"
    if token is None:
        token = settings.access_token

    if token is None:
        raise InvalidTokenError("No token found for workspace")

    if settings.debug:
        click.echo("")
        click.echo("Graphql Query:")
        click.echo(f"URL: {url}")
        click.echo(f"Query: {query}")
        click.echo(f"Variables: {variables}")

    session = create_requests_session()

    response = session.post(
        url,
        headers={
            "User-Agent": f"openhexa-cli/{version('openhexa.sdk')}",
            "Authorization": f"Bearer {token}",
        },
        json={"query": query, "variables": variables},
    )
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise GraphQLError(str(e))

    data = response.json()

    if settings.debug:
        click.echo("Graphql Response:")
        click.echo(data)
        click.echo("")

    if data.get("errors"):
        if data.get("errors")[0].get("extensions", {}).get("code") == "UNAUTHENTICATED":
            raise InvalidTokenError
        raise GraphQLError(data["errors"])
    return data["data"]


def get_skeleton_dir():
    """Get the path to the skeleton directory."""
    return Path(__file__).parent / "skeleton"


def get_workspace(slug: str, token: str):
    """Get a single workspace."""
    return graphql(
        """
            query getWorkspace($slug: String!) {
                workspace(slug: $slug) {
                    name
                    slug
                }
            }
            """,
        {"slug": slug},
        token,
    )["workspace"]


def get_pipelines_pages(name=None):
    """Get the first page of pipelines in the workspace optionally ranked by name similarity."""
    if settings.current_workspace is None:
        raise NoActiveWorkspaceError
    data = graphql(
        """
    query getWorkspacePipelines($workspaceSlug: String!, $name: String, $page: Int = 1, $perPage: Int = 10) {
        pipelines(workspaceSlug: $workspaceSlug, name: $name, page: $page, perPage: $perPage) {
            totalPages
            items {
                id
                code
                name
                type
                currentVersion {
                    id
                    name
                }
            }
        }
    }
    """,
        {"workspaceSlug": settings.current_workspace, "name": name},
    )
    return data["pipelines"]


def get_pipelines(name=None):
    """Get pipelines in the workspace optionally ranked by name similarity."""
    return get_pipelines_pages(name)["items"]


def get_pipeline_from_code(pipeline_code: str) -> dict[str, typing.Any]:
    """Get a single pipeline."""
    if settings.current_workspace is None:
        raise NoActiveWorkspaceError
    data = graphql(
        """
    query getWorkspacePipeline($workspaceSlug: String!, $pipelineCode: String!) {
        pipelineByCode (workspaceSlug: $workspaceSlug, code: $pipelineCode) {
            id
            code
            currentVersion {
                id
                name
            }
        }
    }
    """,
        {
            "workspaceSlug": settings.current_workspace,
            "pipelineCode": pipeline_code,
        },
    )
    return data["pipelineByCode"]


def create_pipeline(pipeline_name: str):
    """Create a pipeline using the API."""
    if settings.current_workspace is None:
        raise NoActiveWorkspaceError
    data = graphql(
        """
    mutation createPipeline($input: CreatePipelineInput!) {
        createPipeline(input: $input) {
            success
            errors
            pipeline {
                id
                code
                name
            }
        }
    }
    """,
        {
            "input": {
                "workspaceSlug": settings.current_workspace,
                "name": pipeline_name,
            }
        },
    )

    if not data["createPipeline"]["success"]:
        raise Exception(data["createPipeline"]["errors"])

    return data["createPipeline"]["pipeline"]


def download_pipeline_sourcecode(pipeline_code, output_path: Path = None, force_overwrite=False):
    """Download the source code of a pipeline."""
    if settings.current_workspace is None:
        raise NoActiveWorkspaceError
    if output_path.exists() and not output_path.is_dir():
        raise OutputDirectoryError(f"{output_path.absolute()} is not a directory")
    if output_path.exists() and output_path.is_dir() and any(output_path.iterdir()) and not force_overwrite:
        raise OutputDirectoryError(f"{output_path.absolute()} is not empty")
    r = graphql(
        """
        query getPipelineVersionSourceCode($workspaceSlug: String!, $pipelineCode: String!)  {
            pipelineByCode(workspaceSlug: $workspaceSlug, code: $pipelineCode) {
                currentVersion {
                    zipfile
                }
            }
        }
    """,
        {
            "workspaceSlug": settings.current_workspace,
            "pipelineCode": pipeline_code,
        },
    )
    if r["pipelineByCode"] is None:
        raise Exception(f"No pipeline exists in {settings.current_workspace} with code {pipeline_code}")
    if r["pipelineByCode"]["currentVersion"] is None:
        raise Exception(f"No version found for pipeline {pipeline_code}")

    zip_file = base64.b64decode(r["pipelineByCode"]["currentVersion"]["zipfile"].encode("ascii"))
    with ZipFile(io.BytesIO(zip_file)) as zf:
        zf.extractall(output_path)


def delete_pipeline(pipeline_id: str):
    """Delete a single pipeline."""
    data = graphql(
        """
    mutation deletePipeline($input: DeletePipelineInput!) {
                    deletePipeline(input: $input) {
                        success
                        errors
                    }
                }
    """,
        {"input": {"id": pipeline_id}},
    )

    if not data["deletePipeline"]["success"]:
        if "PERMISSION_DENIED" in data["deletePipeline"]["errors"]:
            raise Exception(
                "Check that you have the correct permission or the pipeline is not in a queued/running state."
            )

        raise Exception(data["deletePipeline"]["errors"])

    return data["deletePipeline"]["success"]


def ensure_is_pipeline_dir(pipeline_path: str):
    """Ensure that there is a pipeline.py file in the directory."""
    if not os.path.isdir(pipeline_path):
        raise PipelineDirectoryError(f"Path {pipeline_path} is not a directory")
    if not os.path.exists(pipeline_path):
        raise PipelineDirectoryError(f"Directory {pipeline_path} does not exist")
    if not os.path.exists(os.path.join(pipeline_path, "pipeline.py")):
        raise PipelineDirectoryError(f"Directory {pipeline_path} does not contain a pipeline.py file")

    return True


def run_pipeline(path: Path, config: dict, image: str = None, debug: bool = False) -> Container:
    """Run a pipeline using the provided configuration."""
    ensure_is_pipeline_dir(path)
    ensure_pipeline_config_exists(path)
    env_vars = get_local_workspace_config(path)
    # # Prepare the mount for the workspace's files
    mount_files_path = Path(env_vars.pop("WORKSPACE_FILES_PATH")).absolute()
    try:
        docker_client = docker.from_env()
        docker_client.ping()
    except docker.errors.DockerException:
        raise DockerError(
            "Docker is not running or is not installed. Please install Docker Desktop and start the service."
        )

    if image is None:
        image = env_vars.get("WORKSPACE_DOCKER_IMAGE", "blsq/openhexa-blsq-environment:latest")

    zip_file = generate_zip_file(path)
    tmp_dir = tempfile.mkdtemp()
    with ZipFile(zip_file) as zf:
        zf.extractall(tmp_dir)

    volumes = {
        tmp_dir: {"bind": "/home/hexa/pipeline", "mode": "rw"},
        mount_files_path: {
            "bind": "/home/hexa/workspace",
            "mode": "rw",
        },
    }

    environment = {
        "HEXA_ENVIRONMENT": "local_pipeline",
        "HEXA_WORKSPACE": settings.current_workspace,
        "REMOTE_DEBUGGER": "true" if debug else None,
        **env_vars,
    }

    command = f"pipeline run --config {base64.b64encode(json.dumps(config).encode('utf-8')).decode('utf-8')}"
    try:
        docker_client.images.get(image)
    except docker.errors.ImageNotFound:
        logging.info("Pulling image %s...", image)
        docker_client.images.pull(image)
        logging.info("Image %s pulled", image)
    try:
        logging.info(f"Creating pipeline container with image '{image}'...")
        return docker_client.containers.run(
            image,
            command,
            remove=True,
            auto_remove=True,
            tty=debug,
            stdin_open=True,
            platform="linux/amd64",
            volumes=volumes,
            ports={"5678": 5678} if debug else None,
            environment=environment,
            healthcheck={"test": ["NONE"]},  # Disable health checks
            detach=True,
        )
    except docker.errors.ContainerError as e:
        raise DockerError(f"Error while running the pipeline: {e}")
    except docker.errors.ImageNotFound:
        raise DockerError("Docker image not found")


# This is easier to mock in the tests than trying to mock click.confirm
def ask_pipeline_config_creation():
    """Mockable function to ask the user if he wants to create a pipeline config file.

    Returns
    -------
        bool: True if the user wants to create a pipeline config file, False otherwise.

    """
    return click.confirm(
        "No workspace.yaml file found. Do you want to create one?",
        default=True,
    )


def ensure_pipeline_config_exists(pipeline_path: Path):
    """Ensure that there is a workspace.yaml file in the directory. If it does not exist, it asks the user if he wants to create it.

    Args:
        pipeline_path (Path): Base directory of the pipeline
    """
    if (pipeline_path / "workspace.yaml").exists():
        return True

    if ask_pipeline_config_creation():
        content = open(get_skeleton_dir() / "workspace.yaml").read()
        with open(pipeline_path / "workspace.yaml", "w") as f:
            f.write(content)
        return True
    else:
        raise Exception("No workspace.yaml file found")


def create_pipeline_structure(pipeline_name: str, base_path: Path, workspace: str, workflow_mode: str = None) -> Path:
    """Create the structure of a pipeline in the provided directory based on a skeleton.

    Args
    -----
        pipeline_name (str): Name of the pipeline.
        base_path (Path): Base directory of the pipeline.
        workspace (str): Slug of the workspace.
        workflow_mode (str): If provided, a GitHub workflow will be created in the pipeline directory. Accepted values are "push", "manual" and "release".

    Returns
    -------
        Path: Path to the created pipeline directory.
    """
    output_directory = base_path / stringcase.snakecase(pipeline_name.lower())

    if output_directory.exists():
        raise ValueError(f"Directory {output_directory} already exists")

    sample_directory_path = get_skeleton_dir()
    templates = [
        "pipeline.py.j2",
        "workspace.yaml",
        ".gitignore",
        ".vscode/launch.json.j2",
    ]
    if workflow_mode is not None:
        templates.append(".github/workflows/push-pipeline.yml.j2")

    vars = {
        "pipeline_code": stringcase.spinalcase(pipeline_name.lower()),
        "pipeline_human_name": pipeline_name,
        "pipeline_snake_name": stringcase.snakecase(pipeline_name.lower()),
        "workspace_slug": workspace,
        "workflow_mode": workflow_mode,
    }
    jinja_options = {
        "variable_start_string": "[[",
        "variable_end_string": "]]",
    }

    for tpl_file in templates:
        out_path = output_directory / Path(tpl_file.rstrip(".j2"))
        template = Template(open(sample_directory_path / Path(tpl_file)).read(), **jinja_options)
        out_path.parent.mkdir(parents=True, exist_ok=True)
        with open(out_path, "w") as f:
            f.write(template.render(**vars))

    (output_directory / "workspace").mkdir(parents=True, exist_ok=True)
    return output_directory


def generate_zip_file(pipeline_directory_path: str | Path) -> io.BytesIO:
    """Generate a ZIP file containing the pipeline code.

    Args:
        pipeline_directory_path (str | Path): The path to the pipeline directory.

    Returns
    -------
        io.BytesIO: A BytesIO object containing the ZIP file.
    """
    if settings.debug:
        click.echo("Generating ZIP file:")
    zip_file = io.BytesIO(b"")
    files = []

    # We exclude the workspace directory since it can break the mount of the bucket on /home/hexa/workspace
    # This is also the default value of the WORKSPACE_FILES_PATH env var
    excluded_paths = [pipeline_directory_path / "workspace"]
    try:
        env_vars = get_local_workspace_config(pipeline_directory_path)
        if env_vars.get("WORKSPACE_FILES_PATH") and Path(env_vars["WORKSPACE_FILES_PATH"]) not in excluded_paths:
            excluded_paths.append(Path(env_vars["WORKSPACE_FILES_PATH"]))
    except FileNotFoundError:
        # No workspace.yaml file found, we can ignore this error and assume the default value of WORKSPACE_FILES_PATH
        pass
    with ZipFile(zip_file, "w") as zipObj:
        for path in pipeline_directory_path.glob("**/*"):
            if path.name == "python":
                # We are in a virtual environment
                excluded_paths.append(path.parent.parent)  # ./<venv>/bin/python -> ./<venv>

            if path.suffix not in (".py", ".ipynb", ".txt", ".md", ".R", ".r"):
                continue

            files.append(path)

        if settings.debug:
            click.echo(f"Excluded dirs: {[p.absolute() for p in excluded_paths]}")

        for file_path in files:
            # Do not include files from the excluded paths
            if any([file_path.is_relative_to(excluded_dir) for excluded_dir in excluded_paths]):
                if settings.debug:
                    click.echo(f"\t{file_path.name} (excluded)")
                continue
            if settings.debug:
                click.echo(f"\t{file_path.name}")
            zipObj.write(file_path, file_path.relative_to(pipeline_directory_path))
    zip_file.seek(0)
    return zip_file


def upload_pipeline(
    target_pipeline_code: str,
    pipeline_directory_path: str | Path,
    name: str = None,
    description: str = None,
    link: str = None,
):
    """Upload the pipeline contained in the provided directory using the GraphQL API.

    The pipeline code will be zipped and base64-encoded before being sent to the backend.
    The target pipeline will be updated with the new version.
    """
    if settings.current_workspace is None:
        raise NoActiveWorkspaceError

    directory = pipeline_directory_path.absolute()
    pipeline = get_pipeline(directory)
    zip_file = generate_zip_file(directory)

    if settings.debug:
        # Write zip_file to disk for debugging
        with open("pipeline.zip", "wb") as debug_file:
            debug_file.write(zip_file.read())
        zip_file.seek(0)

    base64_content = base64.b64encode(zip_file.read()).decode("ascii")
    data = graphql(
        """
            mutation uploadPipeline($input: UploadPipelineInput!) {
                uploadPipeline(input: $input) {
                    success
                    errors
                    pipelineVersion {
                        id
                        versionName
                        pipeline {
                            id
                            permissions {
                                createTemplateVersion {
                                    isAllowed
                                }
                            }
                            template {
                                id
                                code
                                name
                            }
                        }
                    }
                }
            }
        """,
        {
            "input": {
                "workspaceSlug": settings.current_workspace,
                "code": target_pipeline_code,
                "name": name,
                "description": description,
                "externalLink": link,
                "zipfile": base64_content,
                "parameters": [p.to_dict() for p in pipeline.parameters],
                "timeout": pipeline.timeout,
            }
        },
    )

    if not data["uploadPipeline"]["success"]:
        if PipelineDefinitionErrorCode.PIPELINE_DOES_NOT_SUPPORT_PARAMETERS.value in data["uploadPipeline"]["errors"]:
            raise InvalidDefinitionError(
                "Cannot push a new version : this pipeline has a schedule and the new version cannot be scheduled "
                "(all parameters must be optional or have default values)."
            )
        elif PipelineDefinitionErrorCode.INVALID_TIMEOUT_VALUE.value in data["uploadPipeline"]["errors"]:
            raise InvalidDefinitionError(
                "Timeout value is invalid : ensure that it's no negative and inferior to 12 hours."
            )
        elif PipelineDefinitionErrorCode.DUPLICATE_PIPELINE_VERSION_NAME.value in data["uploadPipeline"]["errors"]:
            raise InvalidDefinitionError("The pipeline version name is already used. Please choose another name.")
        else:
            raise Exception(data["uploadPipeline"]["errors"])

    return data["uploadPipeline"]["pipelineVersion"]


class PipelineTemplateVersionCreateErrorCode(enum.Enum):
    """Enumeration of possible pipeline template version create error codes."""

    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    PIPELINE_VERSION_NOT_FOUND = "PIPELINE_VERSION_NOT_FOUND"


def create_pipeline_template_version(
    workspace_slug: str, pipeline_id: str, pipeline_version_id: str, changelog: str = None
):
    """Create a pipeline template version using the API."""
    data = graphql(
        """
    mutation createPipelineTemplateVersion($input: CreatePipelineTemplateVersionInput!) {
        createPipelineTemplateVersion(input: $input) {
            success
            errors
            pipelineTemplate {
                id
                name
                code
                currentVersion {
                    id
                    versionNumber
                }
            }
        }
    }
    """,
        {
            "input": {
                "workspaceSlug": workspace_slug,
                "pipelineId": pipeline_id,
                "pipelineVersionId": pipeline_version_id,
                "changelog": changelog,
            }
        },
    )
    if not data["createPipelineTemplateVersion"]["success"]:
        if (
            PipelineTemplateVersionCreateErrorCode.PERMISSION_DENIED.value
            in data["createPipelineTemplateVersion"]["errors"]
        ):
            raise PermissionDenied("Permission denied")
        elif (
            PipelineTemplateVersionCreateErrorCode.WORKSPACE_NOT_FOUND.value
            in data["createPipelineTemplateVersion"]["errors"]
        ):
            raise Exception("Workspace not found")
        elif (
            PipelineTemplateVersionCreateErrorCode.PIPELINE_NOT_FOUND.value
            in data["createPipelineTemplateVersion"]["errors"]
        ):
            raise Exception("Pipeline not found")
        elif (
            PipelineTemplateVersionCreateErrorCode.PIPELINE_VERSION_NOT_FOUND.value
            in data["createPipelineTemplateVersion"]["errors"]
        ):
            raise Exception("Pipeline version not found for this pipeline")
        else:
            raise Exception(data["createPipelineTemplateVersion"]["errors"])

    return data["createPipelineTemplateVersion"]["pipelineTemplate"]


def is_dhis2_connection_up(workspace_slug: str, connection_slug: str) -> bool:
    """DHIS2 connection status."""
    response = graphql(
        """
        query getConnectionBySlug($workspaceSlug: String!, $connectionSlug: String!) {
        connectionBySlug(workspaceSlug:$workspaceSlug, connectionSlug: $connectionSlug){
            ... on DHIS2Connection {
                    status
                }
        }
        }
        """,
        variables={
            "workspaceSlug": workspace_slug,
            "connectionSlug": connection_slug,
        },
    )
    return response["data"]["connectionBySlug"]["status"] == "UP"


class OpenHexaClient(BaseOpenHexaClient):
    """OpenHexaClient is a class that provides methods to interact with the OpenHexa GraphQL API."""

    def __init__(self, token=None):
        """Initialize the OpenHexaClient with the OpenHexa API URL and headers."""
        super().__init__(url=settings.api_url + "/graphql/", token=settings.access_token)

    def execute(self, query, **kwargs):
        """Decorate parent execute method to log the GraphQL query and response."""
        _detect_graphql_breaking_changes(token=self.token)

        if self.token is None:
            raise InvalidTokenError("No token found for workspace")

        if settings.debug:
            click.echo("")
            click.echo("Graphql Query:")
            click.echo(f"URL: {self.url}")
            click.echo(f"Query: {query}")
            variables = kwargs.get("variables", {})
            click.echo(f"Variables: {variables}")

        response = super().execute(query=query, **kwargs)

        if settings.debug:
            click.echo("")
            click.echo("Graphql Response:")
            click.echo(f"Response: {response}")

        return response

    def get_data(self, response: httpx.Response) -> dict[str, Any]:
        """Get the data from the response, handling errors and authentication issues."""
        try:
            data = super().get_data(response)
        except Exception as e:
            if "Resolver requires an authenticated user" in str(e):
                raise InvalidTokenError("No or invalid token found for workspace, please check your configuration.")
            raise
        return data



"""Parameter test module."""

from dataclasses import make_dataclass
from unittest import mock

import pytest

from openhexa.sdk import (
    CustomConnection,
    Dataset,
    DHIS2Connection,
    GCSConnection,
    IASOConnection,
    PostgreSQLConnection,
    S3Connection,
    workspace,
)
from openhexa.sdk.pipelines.parameter import (
    Boolean,
    CustomConnectionType,
    DatasetType,
    DHIS2ConnectionType,
    Float,
    FunctionWithParameter,
    GCSConnectionType,
    IASOConnectionType,
    Integer,
    InvalidParameterError,
    Parameter,
    ParameterValueError,
    PostgreSQLConnectionType,
    S3ConnectionType,
    StringType,
    parameter,
)
from openhexa.utils import stringcase


def test_parameter_types_normalize():
    """Check normalization or basic types."""
    # StringType
    string_parameter_type = StringType()
    assert string_parameter_type.normalize("a string") == "a string"
    assert string_parameter_type.normalize(" a string ") == "a string"
    assert string_parameter_type.normalize("") is None
    assert string_parameter_type.normalize(" ") is None

    # Integer
    integer_parameter_type = Integer()
    assert integer_parameter_type.normalize(99) == 99
    assert integer_parameter_type.normalize("abc") == "abc"

    # Float
    float_parameter_type = Float()
    assert float_parameter_type.normalize(3.14) == 3.14
    assert float_parameter_type.normalize(3) == 3.0

    # Boolean
    boolean_parameter_type = Boolean()
    assert boolean_parameter_type.normalize(True) is True
    assert boolean_parameter_type.normalize(False) is False
    assert boolean_parameter_type.normalize(3) == 3


def test_parameter_types_validate():
    """Sanity checks for basic types validation."""
    # StringType
    string_parameter_type = StringType()
    assert string_parameter_type.validate("a string") == "a string"
    with pytest.raises(ParameterValueError):
        string_parameter_type.validate(86)

    # Integer
    integer_parameter_type = Integer()
    assert integer_parameter_type.validate(99) == 99
    with pytest.raises(ParameterValueError):
        integer_parameter_type.validate("not an int")

    # Float
    float_parameter_type = Float()
    assert float_parameter_type.validate(3.14) == 3.14
    with pytest.raises(ParameterValueError):
        float_parameter_type.validate("3.14")

    # Boolean
    boolean_parameter_type = Boolean()
    assert boolean_parameter_type.validate(True) is True
    assert boolean_parameter_type.validate(False) is False
    with pytest.raises(ParameterValueError):
        boolean_parameter_type.validate(86)


def test_validate_postgres_connection():
    """Check PostgreSQL connection validation."""
    identifier = "polio-ff3a0d"

    host = "https://172.17.0.1"
    port = 5432
    username = "dhis2"
    password = "dhis2_pwd"
    database_name = "polio"
    data = PostgreSQLConnection(host, port, username, password, database_name)
    with mock.patch.object(workspace, "postgresql_connection", return_value=data):
        postgres_parameter_type = PostgreSQLConnectionType()
        assert postgres_parameter_type.validate(identifier) == PostgreSQLConnection(
            host, port, username, password, database_name
        )
        with pytest.raises(ParameterValueError):
            postgres_parameter_type.validate(86)


def test_validate_dhis2_connection():
    """Check DHIS2 connection validation."""
    identifier = "dhis2-connection-id"

    url = "https://test.dhis2.org/"
    username = "dhis2"
    password = "dhis2_pwd"

    data = DHIS2Connection(url, username, password)
    with mock.patch.object(workspace, "dhis2_connection", return_value=data):
        dhis2_parameter_type = DHIS2ConnectionType()
        assert dhis2_parameter_type.validate(identifier) == DHIS2Connection(url, username, password)
        with pytest.raises(ParameterValueError):
            dhis2_parameter_type.validate(86)


def test_validate_iaso_connection():
    """Check IASO connection validation."""
    identifier = "iaso-connection-id"

    url = "https://test.iaso.org/"
    username = "iaso"
    password = "iaso_pwd"

    data = IASOConnection(url, username, password)
    with mock.patch.object(workspace, "iaso_connection", return_value=data):
        iaso_parameter_type = IASOConnectionType()
        assert iaso_parameter_type.validate(identifier) == IASOConnection(url, username, password)
        with pytest.raises(ParameterValueError):
            iaso_parameter_type.validate(86)


def test_validate_gcs_connection():
    """Check GCS connection validation."""
    identifier = "gcs-connection-id"
    service_account_key = "HqQBxH0BAI3zF7kANUNlGg"
    bucket_name = "test"

    data = GCSConnection(service_account_key, bucket_name)
    with mock.patch.object(workspace, "gcs_connection", return_value=data):
        gcs_parameter_type = GCSConnectionType()
        assert gcs_parameter_type.validate(identifier) == GCSConnection(service_account_key, bucket_name)
        with pytest.raises(ParameterValueError):
            gcs_parameter_type.validate(86)


def test_validate_s3_connection():
    """Check S3 connection validation."""
    identifier = "s3-connection-id"

    secret_access_key = "HqQBxH0BAI3zF7kANUNlGg"
    access_key_id = "84hVntMaMSYP/RSW9ex04w"
    bucket_name = "test"

    data = S3Connection(access_key_id, secret_access_key, bucket_name)
    with mock.patch.object(workspace, "s3_connection", return_value=data):
        s3_parameter_type = S3ConnectionType()
        assert s3_parameter_type.validate(identifier) == S3Connection(access_key_id, secret_access_key, bucket_name)
        with pytest.raises(ParameterValueError):
            s3_parameter_type.validate(86)


def test_validate_custom_connection(monkeypatch):
    """Check Custom connection validation."""
    identifier = "custom-connection-id"

    field_1 = "field_1"
    field_2 = "field_2"

    dataclass = make_dataclass(
        stringcase.pascalcase(identifier),
        [field_1, field_2],
        bases=(CustomConnection,),
        repr=False,
    )
    monkeypatch.setenv("HEXA_SERVER_URL", "http://app.openhexa.test")
    with mock.patch.object(workspace, "get_connection", return_value=dataclass(field_1, field_2)):
        custom_co_type = CustomConnectionType()

        custom_co = custom_co_type.validate(identifier)
        _custom_co = workspace.custom_connection(identifier)

        assert str(custom_co) == str(_custom_co)
        with pytest.raises(ParameterValueError):
            custom_co_type.validate(86)


@mock.patch("openhexa.sdk.workspace.get_dataset")
def test_validate_dataset_parameter(mock_get_dataset):
    """Check Dataset parameter validation."""
    identifier = "dataset-slug"

    dataset = Dataset(id="id", slug=identifier, name="name", description="Description")

    mock_get_dataset.return_value = dataset

    dataset_type = DatasetType()

    assert dataset_type.validate(identifier) == dataset
    with pytest.raises(ParameterValueError):
        dataset_type.validate(86)


def test_parameter_init():
    """Sanity checks for parameter initialization."""
    # Wrong type
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type="string")  # NOQA

    # Wrong code
    with pytest.raises(InvalidParameterError):
        Parameter("-123", type=str)
    with pytest.raises(InvalidParameterError):
        Parameter("Abc", type=str)
    with pytest.raises(InvalidParameterError):
        Parameter("0_z", type=str)

    # Empty choices
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, choices=[])

    # Invalid choices
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, choices=[1, 2, 3])

    # Boolean can't have choices
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=bool, choices=[True, False])

    # Boolean can't be multiple
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=bool, multiple=True)

    # Invalid defaults
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=bool, default=3)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default=[1, 2], multiple=True)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default="yolo", multiple=True)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default="")
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default=[""], multiple=True)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, choices=["foo", "bar"], default="Bar")


def test_parameter_validate_single():
    """Base check for single-value validation."""
    # required is True by default
    parameter_1 = Parameter("arg1", type=str)
    assert parameter_1.validate("a valid string") == "a valid string"
    with pytest.raises(ParameterValueError):
        parameter_1.validate(None)
    with pytest.raises(ParameterValueError):
        parameter_1.validate("")

    # still required, but a default is provided
    parameter_2 = Parameter("arg2", type=int, default=3)
    assert parameter_2.validate(None) == 3

    # not required, no default
    parameter_3 = Parameter("arg3", type=bool, required=False)
    assert parameter_3.validate(None) is None

    # choices
    parameter_4 = Parameter("arg4", type=str, choices=["ab", "cd"])
    assert parameter_4.validate("ab") == "ab"
    with pytest.raises(ParameterValueError):
        parameter_4.validate("xy")


def test_parameter_validate_multiple():
    """Test multiple values validation rules."""
    # required is True by default
    parameter_1 = Parameter("arg1", type=str, multiple=True)
    assert parameter_1.validate(["Valid string", "Another valid string"]) == [
        "Valid string",
        "Another valid string",
    ]
    with pytest.raises(ParameterValueError):
        parameter_1.validate(None)
    with pytest.raises(ParameterValueError):
        parameter_1.validate([])
    with pytest.raises(ParameterValueError):
        parameter_1.validate(["", ""])

    # still required, but a default is provided
    parameter_2 = Parameter("arg2", type=int, multiple=True, default=[2, 3])
    assert parameter_2.validate(None) == [2, 3]

    # not required, no default
    parameter_3 = Parameter("arg3", type=int, required=False, multiple=True)
    assert parameter_3.validate(None) == []
    assert parameter_3.validate([]) == []

    # choices
    parameter_4 = Parameter("arg4", type=str, default=["ab", "ef"], choices=["ab", "cd", "ef"], multiple=True)
    assert parameter_4.validate(["ab"]) == ["ab"]
    assert parameter_4.validate(["ab", "cd"]) == ["ab", "cd"]
    with pytest.raises(ParameterValueError):
        parameter_4.validate(["xy"])
    with pytest.raises(ParameterValueError):
        parameter_4.validate(["ab", "xy"])


def test_parameter_decorator():
    """Ensure that the @parameter decorator behaves as expected (options and defaults)."""

    @parameter("arg1", type=int)
    @parameter(
        "arg2",
        type=str,
        name="Arg 2",
        help="Help 2",
        default=["yo"],
        required=False,
        multiple=True,
    )
    def a_function():
        pass

    assert isinstance(a_function, FunctionWithParameter)
    function_parameters = a_function.get_all_parameters()
    assert len(function_parameters) == 2
    assert all(isinstance(a, Parameter) for a in function_parameters)

    assert function_parameters[0].code == "arg1"
    assert isinstance(function_parameters[0].type, Integer)
    assert function_parameters[0].name is None
    assert function_parameters[0].help is None
    assert function_parameters[0].default is None
    assert function_parameters[0].required is True
    assert function_parameters[0].multiple is False

    assert function_parameters[1].code == "arg2"
    assert isinstance(function_parameters[1].type, StringType)
    assert function_parameters[1].name == "Arg 2"
    assert function_parameters[1].help == "Help 2"
    assert function_parameters[1].default == ["yo"]
    assert function_parameters[1].required is False
    assert function_parameters[1].multiple is True



"""CLI module, with click."""

import json
import signal
import urllib
from datetime import datetime
from importlib.metadata import version
from pathlib import Path
from urllib.parse import urlparse

import click

from openhexa.cli.api import (
    DockerError,
    InvalidDefinitionError,
    NoActiveWorkspaceError,
    OpenHexaClient,
    OutputDirectoryError,
    PipelineDirectoryError,
    create_pipeline,
    create_pipeline_structure,
    create_pipeline_template_version,
    delete_pipeline,
    download_pipeline_sourcecode,
    ensure_is_pipeline_dir,
    get_library_versions,
    get_pipeline_from_code,
    get_pipelines_pages,
    get_workspace,
    run_pipeline,
    upload_pipeline,
)
from openhexa.cli.settings import settings, setup_logging
from openhexa.sdk.pipelines.exceptions import PipelineNotFound
from openhexa.sdk.pipelines.runtime import get_pipeline


def validate_url(ctx, param, value):
    """Validate URL format."""
    if value is not None:
        parsed_url = urlparse(value)
        if parsed_url.scheme in ("http", "https") and parsed_url.netloc:
            return value
        else:
            raise click.BadParameter("Invalid URL format. Please provide a valid HTTP or HTTPS URL.")


@click.group()
@click.version_option(version("openhexa.sdk"))
@click.pass_context
def app(ctx):
    """OpenHEXA CLI."""
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    setup_logging()
    ctx.ensure_object(dict)

    # Check if the version is outdated and warns the user if it's the case
    ONE_HOUR = 60 * 60
    now_timestamp = int(datetime.now().timestamp())
    if settings.last_version_check is None or now_timestamp - settings.last_version_check > ONE_HOUR:
        installed_version, latest_version = get_library_versions()
        settings.last_version_check = now_timestamp
        if installed_version != latest_version or True:
            click.secho(
                "\n".join(
                    (
                        "╔════════════════════════════════════════════════════════════════════════════════════════╗",
                        f"║ 🚨 Your OpenHEXA CLI version is outdated. Please update to the latest version ({latest_version}) ║",
                        "║ $ pip install --upgrade openhexa.sdk                                                   ║",
                        "╚════════════════════════════════════════════════════════════════════════════════════════╝",
                        "",
                    )
                ),
                fg="yellow",
            )


@app.group(invoke_without_command=True)
@click.pass_context
def workspaces(ctx):
    """Manage workspaces (add workspace, remove workspace, list workspaces, activate a workspace)."""
    if ctx.invoked_subcommand is None:
        ctx.forward(workspaces_list)


@workspaces.command(name="add")
@click.argument("slug")
@click.option(
    "--token",
    prompt=True,
    hide_input=True,
    confirmation_prompt=False,
    envvar="HEXA_TOKEN",
)
def workspaces_add(slug, token):
    """Add a workspace to the configuration and activate it. The access token is required to access the workspace."""
    if slug in settings.workspaces:
        click.echo(f"Workspace {slug} already exists. We will only update its token.")
    else:
        click.echo(f"Adding workspace {slug}")
    if get_workspace(slug, token):
        settings.add_workspace(slug, token)
    else:
        _terminate(
            f"Workspace {slug} does not exist on {settings.api_url}.",
            err=True,
        )


@workspaces.command(name="activate")
@click.argument("slug")
def workspaces_activate(slug):
    """Activate a workspace that is already in the configuration. The activated workspace will be used for the 'pipelines' commands."""
    try:
        settings.activate(slug)
        click.echo(f"Current workspace set to {click.style(slug, bold=True)}")
    except ValueError:
        _terminate(
            f"Workspace {slug} does not exist on {settings.api_url}. Available workspaces:\n {', '.join(settings.workspaces.keys())}"
        )


@workspaces.command(name="list")
def workspaces_list():
    """List the workspaces in the configuration."""
    click.echo("Workspaces:")
    for slug in settings.workspaces.keys():
        click.echo(click.style(f"* {slug} (active)", bold=True) if slug == settings.current_workspace else f"* {slug}")


@workspaces.command(name="rm")
@click.argument("slug")
def workspaces_rm(slug):
    """Remove a workspace from the configuration.

    SLUG is the slug of the workspace to remove from the configuration.
    """
    if slug not in settings.workspaces:
        click.echo(f"Workspace {slug} does not exist")
        return
    click.echo(f"Removing workspace {slug}")

    settings.remove_workspace(slug)


@app.group(invoke_without_command=True)
@click.pass_context
def config(ctx):
    """Manage configuration of the CLI."""
    if ctx.invoked_subcommand is None:
        click.echo("Debug: " + ("True" if settings.debug else "False"))
        click.echo(f"Backend URL: {settings.api_url}")
        click.echo(f"Current workspace: {settings.current_workspace}")
        click.echo("\nWorkspaces:")
        click.echo("\n".join(settings.workspaces.keys()))


@config.command(name="set_url")
@click.argument("url")
def config_set_url(url):
    """Set the URL of the backend."""
    settings.set_api_url(url)
    click.echo(f"Set backend URL to {settings.api_url}")


@app.group(invoke_without_command=True)
@click.pass_context
def pipelines(ctx):
    """Manage pipelines (list pipelines, push a pipeline to the backend)."""
    if ctx.invoked_subcommand is None:
        ctx.forward(pipelines_list)


@pipelines.command("init")
@click.argument("name", type=str)
def pipelines_init(name: str):
    """Initialize a new pipeline in a fresh directory."""
    click.echo(f"Creating pipeline {name}...")
    workflow_mode = None
    modes = [
        ("push", "Push the pipeline on new commits on main"),
        ("release", "Push the pipeline when a git tag is created"),
        ("manual", "Push the pipeline manually from GitHub actions"),
    ]
    if click.confirm(
        "Do you want to create a workflow to publish the pipeline to OpenHEXA from GitHub?",
        default=True,
    ):
        if settings.current_workspace is None:
            _terminate(
                "No workspace activated. Use openhexa workspaces add or openhexa workspaces activate to "
                "activate a workspace.",
                err=True,
            )

        click.echo("Select the trigger for the GitHub action:")
        for idx, (mode, text) in enumerate(modes):
            click.echo(f" {click.style(idx + 1, bold=True)}: ({click.style(mode, bold=True)}) {text}")
        mode_choice = click.prompt("Select a mode", type=click.IntRange(min=1, max=3), default=1)
        workflow_mode = modes[mode_choice - 1][0]
    try:
        pipeline_directory = create_pipeline_structure(
            name,
            Path.cwd(),
            workspace=settings.current_workspace,
            workflow_mode=workflow_mode,
        )

        if workflow_mode:
            click.echo(
                f"Github workflow created. Please add your OpenHEXA pipeline token to the secrets of your repository with the key '{click.style('OH_TOKEN', bold=True)}'."
            )
    except ValueError:
        _terminate(
            click.style(
                f"\n❌ Error while creating pipeline {name}. The directory already exists. Please choose a different name for your pipeline.",
                fg="red",
            ),
            err=True,
        )
    else:
        # Success
        click.echo("")
        click.echo(
            f"{click.style('✅ Success!', fg='green')} Your pipeline has been created in {click.style(pipeline_directory, underline=True, fg='bright_blue')}"
        )


def propose_to_create_template_version(workspace, pipeline_version, yes):
    """Propose to create a new version of the template if the pipeline is based on a template."""
    pipeline = pipeline_version["pipeline"]
    if pipeline["template"] and pipeline["permissions"]["createTemplateVersion"]["isAllowed"]:
        if not yes and not click.confirm(
            f"The template {click.style(pipeline['template']['name'], bold=True)} is based on this pipeline, do you want to publish a new version of the template as well?",
            True,
        ):
            # Return early when the user do not want to create a template version
            return
        changelog = (
            ""
            if yes
            else click.prompt(
                f"{click.style('Changelog', bold=True)} (optional)", type=str, default="", show_default=False
            )
        )
        try:
            template = create_pipeline_template_version(workspace, pipeline["id"], pipeline_version["id"], changelog)
            template_code_url = urllib.parse.quote(template["code"])
            click.echo(
                click.style(
                    f"✅ New version '{template['currentVersion']['versionNumber']}' of the template '{template['name']}' created! You can view the new template version in OpenHEXA on {click.style(f'{settings.public_api_url}/workspaces/{workspace}/templates/{template_code_url}/versions', fg='bright_blue', underline=True)}",
                    fg="green",
                )
            )
        except Exception as e:
            _terminate(
                f'❌ Error while creating template version: "{e}"',
                err=True,
                exception=e,
            )


def select_pipeline(workspace_pipelines, number_of_pages: int, pipeline):
    """Select a pipeline from the list of workspace pipelines or select creating a new one or select a pipeline from a code."""
    create_new_pipeline = f"Create a new {click.style(pipeline.name, bold=True)} pipeline"
    enter_pipeline_code = f"Insert a {click.style('pipeline code', italic=True)}"
    cancel = "Cancel"

    def _generate_choices():
        """Generate the list of choices for the user."""
        return (
            [
                f"{click.style(p['name'], bold=True)} (code: {click.style(p['code'], italic=True)})"
                for p in workspace_pipelines
            ]
            + [create_new_pipeline]
            + ([enter_pipeline_code] if number_of_pages > 1 else [])
            + [cancel]
        )

    def _handle_user_selection(choices):
        """Handle the user's selection from the list of choices."""
        click.echo("Which pipeline do you want to update?")
        for index, choice in enumerate(choices, start=1):
            click.echo(f"[{index}] {choice}")

        choice_idx = (
            click.prompt(
                "Select an option",
                type=click.IntRange(1, len(choices)),
                default=1,
            )
            - 1
        )

        if choices[choice_idx] == create_new_pipeline:
            return None
        elif choices[choice_idx] == enter_pipeline_code:
            return _handle_enter_pipeline_code()
        elif choices[choice_idx] == cancel:
            raise click.Abort()
        else:
            return workspace_pipelines[choice_idx]

    def _handle_enter_pipeline_code():
        """Handle the case where the user wants to enter a pipeline code."""
        pipeline_code = click.prompt(enter_pipeline_code)
        selected_pipeline = get_pipeline_from_code(pipeline_code)
        if selected_pipeline:
            return selected_pipeline
        else:
            click.echo(f"Pipeline with code {click.style(pipeline_code, italic=True)} not found. Please try again.")
            return _handle_enter_pipeline_code()

    return _handle_user_selection(_generate_choices())


@pipelines.command("push")
@click.argument(
    "path",
    type=click.Path(
        exists=True,
        path_type=Path,
        file_okay=False,
        dir_okay=True,
    ),
)
@click.option(
    "--code",
    "-c",
    type=str,
    help="Code of the pipeline",
    prompt="Code of the pipeline",
    prompt_required=False,
)
@click.option(
    "--name",
    "-n",
    default=None,
    type=str,
    help="Name of the version",
    prompt="Name of the version",
    prompt_required=False,
)
@click.option(
    "--description",
    "-d",
    default=None,
    type=str,
    help="Description of the version",
    prompt="Description of the version",
    prompt_required=False,
)
@click.option(
    "--link",
    "-l",
    type=str,
    callback=validate_url,
    help="Link to the version commit",
    prompt="Link of the version release",
    prompt_required=False,
)
@click.option("--yes", is_flag=True, help="Skip confirmation")
def pipelines_push(
    path: str,
    code: str = None,
    name: str = None,
    description: str = None,
    link: str = None,
    yes: bool = False,
):
    """Push a pipeline to the backend. If the pipeline already exists, it will be updated otherwise it will be created.

    PATH is the path to the pipeline file.
    """
    workspace = settings.current_workspace
    if workspace is None:
        _terminate(
            "❌ No workspace activated. Use openhexa workspaces add or openhexa workspaces activate to "
            "activate a workspace.",
            err=True,
        )
    if yes and not code:
        _terminate("❌ You must provide a pipeline code (using -c or --code) when using the --yes flag.", err=True)

    ensure_is_pipeline_dir(path)

    try:
        pipeline = get_pipeline(path)
    except PipelineNotFound:
        _terminate(
            f"❌ No function with openhexa.sdk pipeline decorator found in {click.style(path, bold=True)}.",
            err=True,
        )
    except Exception as e:
        _terminate(f'❌ Error while importing pipeline: "{e}"', exception=e, err=True)
    else:
        pipeline_pages = get_pipelines_pages(name=pipeline.name)
        workspace_pipelines = pipeline_pages["items"]
        number_of_pages = pipeline_pages["totalPages"]
        if settings.debug:
            click.echo(workspace_pipelines)

        if code:
            selected_pipeline = get_pipeline_from_code(code) or _terminate(
                f"❌ Pipeline with code '{code}' not found.", err=True
            )
        elif yes:
            selected_pipeline = workspace_pipelines[0] if workspace_pipelines else None
        else:
            selected_pipeline = select_pipeline(workspace_pipelines, number_of_pages, pipeline)

        if not yes:
            name_text = f" with name {click.style(name, bold=True)}" if name else ""
            confirmation_message = (
                f"Pushing pipeline {click.style(pipeline.name, bold=True)} "
                f"to workspace {click.style(workspace, bold=True)}{name_text} ?"
            )
            click.confirm(confirmation_message, default=True, abort=True)

        selected_pipeline = selected_pipeline or create_pipeline(pipeline.name)
        uploaded_pipeline_version = None
        try:
            uploaded_pipeline_version = upload_pipeline(
                selected_pipeline["code"], path, name, description=description, link=link
            )
            version_url = click.style(
                f"{settings.public_api_url}/workspaces/{workspace}/pipelines/{selected_pipeline['code']}",
                fg="bright_blue",
                underline=True,
            )
            click.echo(
                click.style(
                    f"✅ New version '{uploaded_pipeline_version['versionName']}' created! You can view the pipeline in OpenHEXA on {version_url}",
                    fg="green",
                )
            )
        except InvalidDefinitionError as e:
            _terminate(
                f'❌ Pipeline definition is invalid: "{e}"',
                err=True,
                exception=e,
            )
        except Exception as e:
            _terminate(
                f'❌ Error while importing pipeline: "{e}"',
                err=True,
                exception=e,
            )
        propose_to_create_template_version(workspace, uploaded_pipeline_version, yes)


@pipelines.command("download")
@click.argument("code", type=str)
@click.argument(
    "output",
    type=click.Path(path_type=Path),
)
def pipelines_download(code: str, output: Path):
    """Download a pipeline and unzip it at the output path given."""
    try:
        force_overwrite = False
        if output.exists() and output.is_dir() and any(output.iterdir()):
            force_overwrite = click.confirm(
                f"Directory {click.style(output.absolute(), bold=True)} is not empty. Overwrite the files?"
            )
        download_pipeline_sourcecode(code, output, force_overwrite)
        click.echo(f"You can find the pipeline source code in '{click.style(output.absolute(), bold=True)}'")
    except OutputDirectoryError as e:
        _terminate(str(e), err=True, exception=e)
    except NoActiveWorkspaceError:
        _terminate(
            "❌ No workspace activated. Use openhexa workspaces add or openhexa workspaces activate to "
            "activate a workspace.",
            err=True,
        )


@pipelines.command("delete")
@click.argument("code", type=str)
def pipelines_delete(code: str):
    """Delete a pipeline and all his versions."""
    if settings.current_workspace is None:
        _terminate(
            "❌ No workspace activated. Use openhexa workspaces add or openhexa workspaces activate to "
            "activate a workspace.",
            err=True,
        )
    else:
        pipeline = get_pipeline_from_code(code)
        if pipeline is None:
            _terminate(
                f"❌  Pipeline {click.style(code, bold=True)} does not exist in workspace {click.style(settings.current_workspace, bold=True)}"
            )

        confirmation_code = click.prompt(
            f'This will remove the pipeline "{click.style(code, bold=True)}" from the "{click.style(settings.current_workspace, bold=True)} workspace. This operation cannot be undone.\nPlease enter "{click.style(code, bold=True)}" to confirm',
            type=str,
        )

        if confirmation_code != code:
            _terminate(
                "❌ Pipeline code and confirmation are different, aborted.",
                err=True,
            )

        try:
            if delete_pipeline(pipeline["id"]):
                click.echo(
                    click.style(
                        f"✅ Pipeline {click.style(code, bold=True)} deleted.",
                        fg="green",
                    )
                )

        except Exception as e:
            _terminate(
                f'❌ Error while deleting pipeline: "{e}"',
                err=True,
                exception=e,
            )


@pipelines.command("run")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
)
@click.option("-c", "config_str", type=str, help="Configuration JSON as a string")
@click.option(
    "-f",
    "config_file",
    type=click.File("r"),
    default=None,
    help="Configuration JSON file",
)
@click.option(
    "--image",
    type=str,
    help="Docker image to use",
)
@click.option("--debug", "-d", is_flag=True, help="Run the pipeline in debug mode (with debugpy)")
def pipelines_run(
    path: str,
    image: str = None,
    config_str: str = "{}",
    config_file: click.File = None,
    debug: bool = False,
):
    """Run a pipeline locally."""
    if config_str and config_file:
        _terminate("❌ You can't specify both -c and -f", err=True)
    config = None
    try:
        if config_file:
            config = json.loads(config_file.read(), strict=False)
        else:
            config = json.loads(config_str or "{}", strict=False)

        container = run_pipeline(path, config, image, debug=debug)
        # Listen to ctrl+c to stop the container
        signal.signal(signal.SIGINT, lambda _, __: container.kill())

        if debug:
            if not (Path(path) / ".vscode" / "launch.json").exists():
                click.secho("\nA launch.json is required in order to debug the pipeline", fg="red")

            click.secho("Open your pipeline directory with VSCode and start the debugging session")
            click.secho(
                "Visit the wiki for more information: https://github.com/BLSQ/openhexa/wiki/Writing-OpenHEXA-pipelines#debugging-and-troubleshooting-your-pipelines",
                fg="blue",
            )

        click.secho("\nRun logs", underline=True)
        for line in container.logs(stream=True):
            click.echo(line.decode("utf-8"))

        result = container.wait()
        click.echo()
        if result["StatusCode"] != 0:
            _terminate("❌ Error in pipeline", err=True)
        else:
            click.echo(click.style("✅ Pipeline finished successfully", fg="green"))
    except json.JSONDecodeError as e:
        _terminate(f"❌ Error while parsing JSON: {e}", err=True, exception=e)
    except PipelineDirectoryError:
        _terminate(f"❌ No pipeline found in {path}", err=True)
    except DockerError as e:
        _terminate(f"❌ Error while running pipeline: {e}", err=True, exception=e)


@pipelines.command("list")
def pipelines_list():
    """List all the remote pipelines of the current workspace."""
    if settings.current_workspace is None:
        _terminate("No workspace activated", err=True)

    workspace_pipelines = OpenHexaClient().pipelines(workspace_slug=settings.current_workspace).items
    if len(workspace_pipelines) == 0:
        click.echo(f"No pipelines in workspace {settings.current_workspace}")
        return
    click.echo("Pipelines:")
    for pipeline in workspace_pipelines:
        if pipeline.type == "zipFile":
            current_version = f"v{pipeline.current_version.version_number}" if pipeline.current_version else "N/A"
        else:
            current_version = "Jupyter notebook"
        click.echo(f"* {pipeline.code} - {pipeline.name} ({current_version})")


def _terminate(message: str, exception: Exception = None, err: bool = False):
    click.echo(click.style(message, fg="red"), err=err)
    if settings.debug and exception:
        raise exception
    raise click.Abort()



"""Log levels for the pipeline runs."""

from enum import IntEnum


class LogLevel(IntEnum):
    """
    Enum representing different log levels.

     - Attributes:
        DEBUG (int): Debug level, value 0.
        INFO (int): Info level, value 1.
        WARNING (int): Warning level, value 2.
        ERROR (int): Error level, value 3.
        CRITICAL (int): Critical level, value 4.

    """

    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3
    CRITICAL = 4

    @classmethod
    def parse_log_level(cls, value) -> "LogLevel":
        """Parse a log level from a string or integer."""
        if isinstance(value, int) and 0 <= value <= 4:
            return LogLevel(value)
        if isinstance(value, str):
            if value.isdigit():
                return cls.parse_log_level(int(value))
            value = value.upper()
            if hasattr(cls, value):
                return getattr(cls, value)
        return cls.INFO



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import UpgradePipelineVersionFromTemplateError


class UpgradePipelineVersionFromTemplate(BaseModel):
    upgrade_pipeline_version_from_template: (
        "UpgradePipelineVersionFromTemplateUpgradePipelineVersionFromTemplate"
    ) = Field(alias="upgradePipelineVersionFromTemplate")


class UpgradePipelineVersionFromTemplateUpgradePipelineVersionFromTemplate(BaseModel):
    success: bool
    errors: List[UpgradePipelineVersionFromTemplateError]


UpgradePipelineVersionFromTemplate.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import DeleteDatasetError


class DeleteDataset(BaseModel):
    delete_dataset: "DeleteDatasetDeleteDataset" = Field(alias="deleteDataset")


class DeleteDatasetDeleteDataset(BaseModel):
    success: bool
    errors: List[DeleteDatasetError]


DeleteDataset.model_rebuild()



import time
from unittest import TestCase, mock

from openhexa.cli.api import _detect_graphql_breaking_changes, graphql


class TestGraphQLFunctions(TestCase):
    @mock.patch("openhexa.cli.api._query_graphql")
    @mock.patch("openhexa.cli.api.get_library_versions")
    def test_detect_graphql_breaking_changes_with_mocked_server_schema(
        self, mock_get_library_versions, mock_query_graphql
    ):
        """Test detect_graphql_breaking_changes with a mocked server schema."""
        mock_get_library_versions.return_value = ["1.2.3", "1000.1.2"]
        mock_query_graphql.return_value = {
            "__schema": {
                "queryType": {"name": "Query"},
                "mutationType": None,
                "subscriptionType": None,
                "types": [
                    {
                        "kind": "OBJECT",
                        "name": "Query",
                        "fields": [
                            {
                                "name": "testField",
                                "args": [],
                                "type": {"kind": "SCALAR", "name": "String"},
                                "isDeprecated": False,
                                "deprecationReason": None,
                            }
                        ],
                        "interfaces": [],
                    },
                    {
                        "kind": "SCALAR",
                        "name": "String",
                    },
                ],
                "directives": [],
            }
        }
        stored_schema = """
        type Query {
            testField: Int
        }
        """
        with mock.patch("pathlib.Path.open", mock.mock_open(read_data=stored_schema)):
            with mock.patch("click.secho") as mock_click_secho:
                _detect_graphql_breaking_changes("test_token")
                mock_click_secho.assert_any_call(
                    "⚠️ Breaking changes detected between the SDK (version 1.2.3) and the server:", fg="red"
                )
                mock_click_secho.assert_any_call("- Query.testField changed type from Int to String.", fg="yellow")

    @mock.patch("openhexa.cli.api._query_graphql")
    @mock.patch("openhexa.cli.api._detect_graphql_breaking_changes")
    def test_graphql(self, mock_detect_graphql_breaking_changes, mock_query_graphql):
        """Test that the graphql function is caching the breaking change detection for 1 hour."""
        with mock.patch("openhexa.cli.api.settings") as mock_settings:
            mock_settings.last_breaking_change_check = time.time() - 59 * 60  # Last checked 59 minutes ago
            mock_query_graphql.return_value = {"data": "response"}

            response = graphql("query", token="test_token")
            mock_detect_graphql_breaking_changes.assert_not_called()
            mock_query_graphql.assert_called_once_with("query", None, "test_token")
            self.assertEqual(response, {"data": "response"})

            time_in_the_past = time.time() - 61 * 60  # Last checked 61 minutes ago
            mock_settings.last_breaking_change_check = time_in_the_past

            response = graphql("query", token="test_token")
            mock_detect_graphql_breaking_changes.assert_called_once_with("test_token")
            self.assertGreater(mock_settings.last_breaking_change_check, time_in_the_past)
            self.assertEqual(response, {"data": "response"})



"""Connection test module."""

import dataclasses


@dataclasses.dataclass
class Connection:
    """Abstract base class for connections."""

    pass


@dataclasses.dataclass
class DHIS2Connection(Connection):
    """DHIS2 connection.

    See https://docs.dhis2.org/ for more information.
    """

    url: str
    username: str
    password: str

    def __repr__(self):
        """Safe representation of the DHIS2 connection (no credentials)."""
        return f"DHIS2Connection(url='{self.url}', username='{self.username}')"


@dataclasses.dataclass
class PostgreSQLConnection(Connection):
    """PostgreSQL database connection.

    See https://www.postgresql.org/docs/ for more information.
    """

    host: str
    port: int
    username: str
    password: str
    database_name: str

    def __repr__(self):
        """Safe representation of the PostgreSQL connection (no credentials)."""
        return (
            f"PostgreSQLConnection(host='{self.host}', port='{self.port}', username='{self.username}', "
            f"database_name='{self.database_name}')"
        )

    @property
    def url(self):
        """Provide a URL to the PostgreSQL database.

        The URL follows the official PostgreSQL specification.
        (See https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING for more information)
        """
        return f"postgresql://{self.username}:{self.password}" f"@{self.host}:{self.port}/{self.database_name}"


@dataclasses.dataclass
class S3Connection(Connection):
    """AWS S3 connection.

    See https://docs.aws.amazon.com/s3/ for more information.
    """

    access_key_id: str
    secret_access_key: str
    bucket_name: str

    def __repr__(self):
        """Safe representation of the S3 connection (no credentials)."""
        return f"S3Connection(bucket_name='{self.bucket_name}')"


@dataclasses.dataclass
class GCSConnection(Connection):
    """Google Cloud Storage connection.

    See https://cloud.google.com/storage/docs for more information.
    """

    service_account_key: str
    bucket_name: str

    def __repr__(self):
        """Safe representation of the GCS connection (no credentials)."""
        return f"GCSConnection(bucket_name='{self.bucket_name}')"


@dataclasses.dataclass
class CustomConnection(Connection):
    """Marker class for custom connections.

    The actual class will be built dynamically through the Workspace.custom_connection() method.
    """

    def __repr__(self):
        """Safe representation of the custom connection (no credentials)."""
        return f"CustomConnection(name='{self.__class__.__name__.lower()}')"


@dataclasses.dataclass
class IASOConnection:
    """IASO connection.

    See https://github.com/BLSQ/iaso for more information.
    """

    url: str
    username: str
    password: str

    def __repr__(self):
        """Safe representation of the IASO connection (no credentials)."""
        return f"IASOConnection(url='{self.url}', username='{self.username}')"


ConnectionClasses = {
    "S3": S3Connection,
    "GCS": GCSConnection,
    "POSTGRESQL": PostgreSQLConnection,
    "DHIS2": DHIS2Connection,
    "IASO": IASOConnection,
    "CUSTOM": CustomConnection,
}



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel


class Datasets(BaseModel):
    datasets: "DatasetsDatasets"


class DatasetsDatasets(BaseModel):
    total_pages: int = Field(alias="totalPages")
    items: List["DatasetsDatasetsItems"]


class DatasetsDatasetsItems(BaseModel):
    id: str
    slug: str
    name: str
    description: Optional[str]
    created_at: Any = Field(alias="createdAt")
    updated_at: Any = Field(alias="updatedAt")
    created_by: Optional["DatasetsDatasetsItemsCreatedBy"] = Field(alias="createdBy")
    permissions: "DatasetsDatasetsItemsPermissions"


class DatasetsDatasetsItemsCreatedBy(BaseModel):
    id: Any
    display_name: str = Field(alias="displayName")


class DatasetsDatasetsItemsPermissions(BaseModel):
    update: bool
    delete: bool


Datasets.model_rebuild()
DatasetsDatasets.model_rebuild()
DatasetsDatasetsItems.model_rebuild()



"""Example pipeline to get data elements from DHIS2."""

from openhexa.sdk import current_run, parameter, pipeline
from openhexa.sdk.pipelines.parameter import DHIS2Widget
from openhexa.sdk.workspaces.connection import DHIS2Connection


@pipeline(name="DHIS2 Data Elements")
@parameter("dhis2_con", type=DHIS2Connection, required=True)
@parameter(
    "data_elements",
    type=str,
    required=True,
    widget=DHIS2Widget.DATA_ELEMENTS,
    multiple=True,
    connection="dhis2_con",
)
def dhis2(dhis2_con, data_elements):
    """Get data elements from DHIS2."""
    print_data_elements(dhis2_con, data_elements)


@dhis2.task
def print_data_elements(dhis2_con, data_elements):
    """Print data elements."""
    current_run.log_info("Printing data elements")

    current_run.log_info(f"Data elements: {data_elements}")


if __name__ == "__main__":
    dhis2()
